# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for updating env vars and other configuration info."""

import copy

from googlecloudsdk.api_lib.run import k8s_object
from googlecloudsdk.api_lib.run import traffic
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.run import config_changes
from googlecloudsdk.command_lib.run import connection_context
from googlecloudsdk.command_lib.run import container_parser
from googlecloudsdk.command_lib.run import exceptions
from googlecloudsdk.command_lib.run import flags
from googlecloudsdk.command_lib.run import iap_util
from googlecloudsdk.command_lib.run import messages_util
from googlecloudsdk.command_lib.run import platforms
from googlecloudsdk.command_lib.run import pretty_print
from googlecloudsdk.command_lib.run import resource_args
from googlecloudsdk.command_lib.run import resource_change_validators
from googlecloudsdk.command_lib.run import serverless_operations
from googlecloudsdk.command_lib.run import stages
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.command_lib.util.concepts import presentation_specs
from googlecloudsdk.core import properties
from googlecloudsdk.core.console import progress_tracker


@base.UniverseCompatible
def ContainerArgGroup(release_track=base.ReleaseTrack.GA):
  """Returns an argument group with all per-container update args."""

  help_text = """
Container Flags

    If the --container or --remove-containers flag is specified the following
    arguments may only be specified after a --container flag.
    """
  group = base.ArgumentGroup(help=help_text)
  group.AddArgument(flags.ImageArg(required=False))
  group.AddArgument(flags.PortArg())
  group.AddArgument(flags.Http2Flag())
  group.AddArgument(flags.MutexEnvVarsFlags(release_track=release_track))
  group.AddArgument(flags.MemoryFlag())
  group.AddArgument(flags.CpuFlag())
  group.AddArgument(flags.WorkdirFlag())
  if release_track != base.ReleaseTrack.GA:
    group.AddArgument(flags.SandboxLauncherFlag())
  group.AddArgument(flags.CommandFlag())
  group.AddArgument(flags.ArgsFlag())
  group.AddArgument(flags.SecretsFlags())
  group.AddArgument(flags.DependsOnFlag())

  group.AddArgument(flags.AddVolumeMountFlag())
  group.AddArgument(flags.RemoveVolumeMountFlag())
  group.AddArgument(flags.ClearVolumeMountsFlag())
  group.AddArgument(flags.StartupProbeFlag())
  group.AddArgument(flags.LivenessProbeFlag())
  group.AddArgument(flags.ReadinessProbeFlag())
  group.AddArgument(flags.GpuFlag())

  return group


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.GA)
class Update(base.Command):
  """Update Cloud Run environment variables and other configuration settings."""

  detailed_help = {
      'DESCRIPTION': (
          """\
          {description}
          """
      ),
      'EXAMPLES': (
          """\
          To update one or more env vars:

              $ {command} myservice --update-env-vars=KEY1=VALUE1,KEY2=VALUE2
         """
      ),
  }

  input_flags = (
      '`--update-env-vars`, `--memory`, `--concurrency`, `--timeout`,'
      ' `--connectivity`, `--image`, `--iap`'
  )

  @classmethod
  def CommonArgs(cls, parser):
    # Flags specific to managed CR
    flags.AddBinAuthzPolicyFlags(parser)
    flags.AddBinAuthzBreakglassFlag(parser)
    flags.AddCloudSQLFlags(parser)
    flags.AddCmekKeyFlag(parser)
    flags.AddCmekKeyRevocationActionTypeFlag(parser)
    flags.AddCpuThrottlingFlag(parser)
    flags.AddCustomAudiencesFlag(parser)
    flags.AddDefaultUrlFlag(parser)
    flags.AddDeployHealthCheckFlag(parser)
    flags.AddEgressSettingsFlag(parser)
    flags.AddEncryptionKeyShutdownHoursFlag(parser)
    flags.AddGpuTypeFlag(parser)
    flags.GpuZonalRedundancyFlag(parser)
    flags.AddRevisionSuffixArg(parser)
    flags.AddSandboxArg(parser)
    flags.AddSessionAffinityFlag(parser)
    flags.AddStartupCpuBoostFlag(parser)
    flags.AddVpcConnectorArgs(parser)
    flags.AddVpcNetworkGroupFlagsForUpdate(parser)
    flags.RemoveContainersFlag().AddToParser(parser)
    flags.AddVolumesFlags(parser, cls.ReleaseTrack())
    flags.AddServiceMinMaxInstancesFlag(parser)
    flags.AddInvokerIamCheckFlag(parser)
    flags.AddScalingFlag(parser)
    # Flags specific to connecting to a cluster
    flags.AddEndpointVisibilityEnum(parser)
    flags.CONFIG_MAP_FLAGS.AddToParser(parser)

    # Flags not specific to any platform
    service_presentation = presentation_specs.ResourcePresentationSpec(
        'SERVICE',
        resource_args.GetServiceResourceSpec(prompt=True),
        'Service to update the configuration of.',
        required=True,
        prefixes=False,
    )
    if cls.ReleaseTrack() != base.ReleaseTrack.GA:
      flags.AddDryRunFlag(parser)
    flags.AddConcurrencyFlag(parser)
    flags.AddTimeoutFlag(parser)
    flags.AddAsyncFlag(parser)
    flags.AddLabelsFlags(parser)
    flags.AddGeneralAnnotationFlags(parser)
    flags.AddMinInstancesFlag(parser)
    flags.AddMaxInstancesFlag(parser)
    flags.AddNoTrafficFlag(parser)
    flags.AddDeployTagFlag(parser)
    flags.AddServiceAccountFlag(parser)
    flags.AddClientNameAndVersionFlags(parser)
    flags.AddIngressFlag(parser)
    flags.AddIapFlag(parser)
    concept_parsers.ConceptParser([service_presentation]).AddToParser(parser)
    # No output by default, can be overridden by --format
    parser.display_info.AddFormat('none')

  @classmethod
  def Args(cls, parser):
    cls.CommonArgs(parser)
    container_args = ContainerArgGroup(cls.ReleaseTrack())
    container_parser.AddContainerFlags(
        parser, container_args, cls.ReleaseTrack()
    )

  def _ConnectionContext(self, args):
    return connection_context.GetConnectionContext(
        args, flags.Product.RUN, self.ReleaseTrack()
    )

  def _AssertChanges(self, changes, flags_text, ignore_empty):
    if ignore_empty:
      return
    if not changes or (
        len(changes) == 1
        and isinstance(
            changes[0],
            config_changes.SetClientNameAndVersionAnnotationChange,
        )
    ):
      raise exceptions.NoConfigurationChangeError(
          'No configuration change requested. '
          'Did you mean to include the flags {}?'.format(flags_text)
      )

  def _GetBaseChanges(self, args, existing_service=None, ignore_empty=False):  # pylint: disable=unused-argument
    changes = flags.GetServiceConfigurationChanges(args, self.ReleaseTrack())
    self._AssertChanges(changes, self.input_flags, ignore_empty)
    changes.insert(
        0,
        config_changes.DeleteAnnotationChange(
            k8s_object.BINAUTHZ_BREAKGLASS_ANNOTATION
        ),
    )
    changes.append(
        config_changes.SetLaunchStageAnnotationChange(self.ReleaseTrack())
    )
    return changes

  def _IsMultiRegion(self):
    return False

  def _GetMultiRegionRegions(
      self, changes, service
  ):  # used by child - pylint: disable=unused-argument
    return None

  def _GetIap(self, args):
    if flags.FlagIsExplicitlySet(args, 'iap'):
      return args.iap
    return None

  def Run(self, args):
    """Update the service resource.

       Different from `deploy` in that it can only update the service spec but
       no IAM or Cloud build changes.

    Args:
      args: Args!

    Returns:
      googlecloudsdk.api_lib.run.Service, the updated service
    """

    conn_context = self._ConnectionContext(args)
    service_ref = args.CONCEPTS.service.Parse()
    flags.ValidateResource(service_ref)
    iap = self._GetIap(args)
    project_id = properties.VALUES.core.project.Get(required=True)

    if iap:
      if iap_util.IsOrglessProject(
          project_id
      ) and not iap_util.IsIapAlreadyEnabled(self.ReleaseTrack()):
        pretty_print.Info(
            '\n {bold}**[Warning]**{reset} Deploying services with IAP'
            ' enabled in a project outside of an Organization and may require'
            ' initial setup via the Cloud Console. Please use the Cloud Run'
            ' UI to enable IAP for the first time in the project. '
            'https://cloud.google.com/run/docs/securing/identity-aware-proxy-cloud-run#custom-oauth-client\n'
        )

    with serverless_operations.Connect(conn_context) as client:
      service = client.GetService(service_ref)
      messages_util.MaybeLogDefaultGpuTypeMessage(args, service)
      changes = self._GetBaseChanges(args, service)
      resource_change_validators.ValidateClearVpcConnector(service, args)
      has_latest = (
          service is None or traffic.LATEST_REVISION_KEY in service.spec_traffic
      )
      creates_revision = config_changes.AdjustsTemplate(changes)
      multiregion_regions = self._GetMultiRegionRegions(changes, service)
      deployment_stages = stages.ServiceStages(
          include_iam_policy_set=False,
          include_route=creates_revision and has_latest,
          include_create_revision=creates_revision,
          regions_list=multiregion_regions,
          include_iap=iap is not None,
      )
      dry_run = getattr(args, 'dry_run', False)
      if dry_run:
        progress_message = 'Validating...'
        failure_message = 'Validation failed'
        result_message = 'validating'
      elif creates_revision:
        progress_message = 'Deploying...'
        failure_message = 'Deployment failed'
        result_message = 'deploying'
      else:
        progress_message = 'Updating...'
        failure_message = 'Update failed'
        result_message = 'updating'

      def _ReleaseService(changes_):
        kwargs = {}
        if dry_run:
          kwargs['dry_run'] = True
        with progress_tracker.StagedProgressTracker(
            progress_message,
            deployment_stages,
            failure_message=failure_message,
            suppress_output=args.async_ or dry_run,
        ) as tracker:
          return client.ReleaseService(
              service_ref,
              changes_,
              self.ReleaseTrack(),
              tracker,
              asyn=args.async_,
              prefetch=service,
              generate_name=(
                  flags.FlagIsExplicitlySet(args, 'revision_suffix')
                  or flags.FlagIsExplicitlySet(args, 'tag')
              ),
              is_verbose=properties.VALUES.core.verbosity.Get() == 'debug',
              iap_enabled=iap,
              multiregion_regions=multiregion_regions,
              **kwargs
          )

      try:
        service = _ReleaseService(changes)
      except exceptions.HttpError as e:
        if flags.ShouldRetryNoZonalRedundancy(args, str(e)):
          changes.append(
              config_changes.GpuZonalRedundancyChange(
                  gpu_zonal_redundancy=False
              )
          )
          service = _ReleaseService(changes)
        else:
          raise e

      if getattr(args, 'dry_run', False):
        pretty_print.Success(
            'Service [{}] has been validated.'.format(service.name)
        )
        return service

      if args.async_:
        if service:
          pretty_print.Success(
              'Service [{{bold}}{serv}{{reset}}] is {result_message} '
              'asynchronously.'.format(
                  serv=service.name, result_message=result_message
              )
          )
        else:
          pretty_print.Info('Service update/deployment may have failed.')
      else:
        if service:
          if creates_revision:
            requires_authentication = not getattr(
                args, 'allow_unauthenticated', True
            )
            if (
                'allow_unauthenticated' not in args
                and conn_context.supports_one_platform
                and platforms.GetPlatform() == platforms.PLATFORM_MANAGED
                and not self._IsMultiRegion()
            ):
              requires_authentication = not client.IsUnauthenticated(
                  service_ref
              )

            project = properties.VALUES.core.project.Get(required=True)
            region = flags.GetRegion(args)
            pretty_print.Success(
                messages_util.GetSuccessMessageForSynchronousDeploy(
                    service,
                    args.no_traffic,
                    show_proxy_message=requires_authentication,
                    project=project,
                    region=region,
                )
            )
          else:
            pretty_print.Success(
                'Service [{{bold}}{serv}{{reset}}] has been updated.'.format(
                    serv=service.name
                )
            )
        else:
          pretty_print.Info('Service update/deployment may have failed.')
      return service


@base.ReleaseTracks(base.ReleaseTrack.BETA)
@base.RegionalEndpointsSupported
class BetaUpdate(Update):
  """Update Cloud Run environment variables and other configuration settings."""

  detailed_help = copy.deepcopy(Update.detailed_help)

  input_flags = (
      '`--update-env-vars`, `--memory`, `--concurrency`, `--timeout`,'
      ' `--connectivity`, `--image`, `--iap`, `--ssh`'
  )

  @classmethod
  def Args(cls, parser):
    cls.CommonArgs(parser)

    flags.AddCpuUtilizationFlag(parser)
    flags.AddConcurrencyUtilizationFlag(parser)
    flags.AddSshFlag(parser)

    # Flags specific to managed CR
    flags.SERVICE_MESH_FLAG.AddToParser(parser)
    container_args = ContainerArgGroup(cls.ReleaseTrack())
    container_parser.AddContainerFlags(
        parser, container_args, cls.ReleaseTrack()
    )


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class AlphaUpdate(BetaUpdate):
  """Update Cloud Run environment variables and other configuration settings."""

  input_flags = (
      '`--update-env-vars`, `--memory`, `--concurrency`, `--timeout`,'
      ' `--connectivity`, `--image`, `--iap`, `--ssh`'
  )

  @classmethod
  def Args(cls, parser):
    cls.CommonArgs(parser)

    # Flags specific to managed CR
    flags.AddRuntimeFlag(parser)
    flags.AddDescriptionFlag(parser)
    flags.SERVICE_MESH_FLAG.AddToParser(parser)
    flags.IDENTITY_FLAG.AddToParser(parser)
    flags.IDENTITY_CERTIFICATE_FLAG.AddToParser(parser)
    flags.IDENTITY_TYPE_FLAG.AddToParser(parser)
    flags.FUNCTIONAL_TYPE_FLAG.AddToParser(parser)
    flags.MESH_DATAPLANE_FLAG.AddToParser(parser)
    flags.AMBIENT_NETWORKING_FLAG.AddToParser(parser)
    flags.AddOverflowScalingFlag(parser)
    flags.AddCpuUtilizationFlag(parser)
    flags.AddConcurrencyUtilizationFlag(parser)
    flags.AddClearPresetFlag(parser)
    flags.AddSshFlag(parser)
    container_args = ContainerArgGroup(cls.ReleaseTrack())
    container_parser.AddContainerFlags(
        parser, container_args, cls.ReleaseTrack()
    )


AlphaUpdate.__doc__ = Update.__doc__



"""Generated message classes for parallelstore version v1alpha.

"""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.protorpclite import messages as _messages
from apitools.base.py import encoding
from apitools.base.py import extra_types


package = 'parallelstore'


class CancelOperationRequest(_messages.Message):
  r"""The request message for Operations.CancelOperation."""


class DestinationGcsBucket(_messages.Message):
  r"""Cloud Storage as the destination of a data transfer.

  Fields:
    uri: Required. URI to a Cloud Storage bucket in the format: `gs:///`. The
      path inside the bucket is optional.
  """

  uri = _messages.StringField(1)


class DestinationParallelstore(_messages.Message):
  r"""Parallelstore as the destination of a data transfer.

  Fields:
    path: Optional. Root directory path to the Paralellstore filesystem,
      starting with `/`. Defaults to `/` if unset.
  """

  path = _messages.StringField(1)


class ExportDataRequest(_messages.Message):
  r"""Export data from Parallelstore to Cloud Storage.

  Fields:
    destinationGcsBucket: Cloud Storage destination.
    metadataOptions: Optional. The metadata options for the export data.
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes since the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
    serviceAccount: Optional. User-specified Service Account (SA) credentials
      to be used when performing the transfer. Use one of the following
      formats: * `{EMAIL_ADDRESS_OR_UNIQUE_ID}` * `projects/{PROJECT_ID_OR_NUM
      BER}/serviceAccounts/{EMAIL_ADDRESS_OR_UNIQUE_ID}` *
      `projects/-/serviceAccounts/{EMAIL_ADDRESS_OR_UNIQUE_ID}` If
      unspecified, the Parallelstore service agent is used: `service-@gcp-sa-
      parallelstore.iam.gserviceaccount.com`
    sourceParallelstore: Parallelstore source.
  """

  destinationGcsBucket = _messages.MessageField('DestinationGcsBucket', 1)
  metadataOptions = _messages.MessageField('TransferMetadataOptions', 2)
  requestId = _messages.StringField(3)
  serviceAccount = _messages.StringField(4)
  sourceParallelstore = _messages.MessageField('SourceParallelstore', 5)


class GoogleProtobufEmpty(_messages.Message):
  r"""A generic empty message that you can re-use to avoid defining duplicated
  empty messages in your APIs. A typical example is to use it as the request
  or the response type of an API method. For instance: service Foo { rpc
  Bar(google.protobuf.Empty) returns (google.protobuf.Empty); }
  """



class ImportDataRequest(_messages.Message):
  r"""Import data from Cloud Storage into a Parallelstore instance.

  Fields:
    destinationParallelstore: Parallelstore destination.
    metadataOptions: Optional. The transfer metadata options for the import
      data.
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes since the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
    serviceAccount: Optional. User-specified service account credentials to be
      used when performing the transfer. Use one of the following formats: *
      `{EMAIL_ADDRESS_OR_UNIQUE_ID}` * `projects/{PROJECT_ID_OR_NUMBER}/servic
      eAccounts/{EMAIL_ADDRESS_OR_UNIQUE_ID}` *
      `projects/-/serviceAccounts/{EMAIL_ADDRESS_OR_UNIQUE_ID}` If
      unspecified, the Parallelstore service agent is used: `service-@gcp-sa-
      parallelstore.iam.gserviceaccount.com`
    sourceGcsBucket: The Cloud Storage source bucket and, optionally, path
      inside the bucket.
  """

  destinationParallelstore = _messages.MessageField('DestinationParallelstore', 1)
  metadataOptions = _messages.MessageField('TransferMetadataOptions', 2)
  requestId = _messages.StringField(3)
  serviceAccount = _messages.StringField(4)
  sourceGcsBucket = _messages.MessageField('SourceGcsBucket', 5)


class Instance(_messages.Message):
  r"""A Parallelstore instance.

  Enums:
    DeploymentTypeValueValuesEnum: Optional. Immutable. The deployment type of
      the instance. Allowed values are: * `SCRATCH`: the instance is a scratch
      instance. * `PERSISTENT`: the instance is a persistent instance.
    DirectoryStripeLevelValueValuesEnum: Optional. Immutable. Stripe level for
      directories. Allowed values are: * `DIRECTORY_STRIPE_LEVEL_MIN`:
      recommended when directories contain a small number of files. *
      `DIRECTORY_STRIPE_LEVEL_BALANCED`: balances performance for workloads
      involving a mix of small and large directories. *
      `DIRECTORY_STRIPE_LEVEL_MAX`: recommended for directories with a large
      number of files.
    FileStripeLevelValueValuesEnum: Optional. Immutable. Stripe level for
      files. Allowed values are: * `FILE_STRIPE_LEVEL_MIN`: offers the best
      performance for small size files. * `FILE_STRIPE_LEVEL_BALANCED`:
      balances performance for workloads involving a mix of small and large
      files. * `FILE_STRIPE_LEVEL_MAX`: higher throughput performance for
      larger files.
    StateValueValuesEnum: Output only. The instance state.

  Messages:
    LabelsValue: Optional. Cloud Labels are a flexible and lightweight
      mechanism for organizing cloud resources into groups that reflect a
      customer's organizational needs and deployment strategies. See
      https://cloud.google.com/resource-manager/docs/labels-overview for
      details.

  Fields:
    accessPoints: Output only. A list of IPv4 addresses used for client side
      configuration.
    capacityGib: Required. Immutable. The instance's storage capacity in
      Gibibytes (GiB). Allowed values are between 12000 and 100000, in
      multiples of 4000; e.g., 12000, 16000, 20000, ...
    createTime: Output only. The time when the instance was created.
    daosVersion: Output only. Deprecated: The version of DAOS software running
      in the instance.
    deploymentType: Optional. Immutable. The deployment type of the instance.
      Allowed values are: * `SCRATCH`: the instance is a scratch instance. *
      `PERSISTENT`: the instance is a persistent instance.
    description: Optional. The description of the instance. 2048 characters or
      less.
    directoryStripeLevel: Optional. Immutable. Stripe level for directories.
      Allowed values are: * `DIRECTORY_STRIPE_LEVEL_MIN`: recommended when
      directories contain a small number of files. *
      `DIRECTORY_STRIPE_LEVEL_BALANCED`: balances performance for workloads
      involving a mix of small and large directories. *
      `DIRECTORY_STRIPE_LEVEL_MAX`: recommended for directories with a large
      number of files.
    effectiveReservedIpRange: Output only. Immutable. The ID of the IP address
      range being used by the instance's VPC network. This field is populated
      by the service and contains the value currently used by the service.
    fileStripeLevel: Optional. Immutable. Stripe level for files. Allowed
      values are: * `FILE_STRIPE_LEVEL_MIN`: offers the best performance for
      small size files. * `FILE_STRIPE_LEVEL_BALANCED`: balances performance
      for workloads involving a mix of small and large files. *
      `FILE_STRIPE_LEVEL_MAX`: higher throughput performance for larger files.
    labels: Optional. Cloud Labels are a flexible and lightweight mechanism
      for organizing cloud resources into groups that reflect a customer's
      organizational needs and deployment strategies. See
      https://cloud.google.com/resource-manager/docs/labels-overview for
      details.
    name: Identifier. The resource name of the instance, in the format
      `projects/{project}/locations/{location}/instances/{instance_id}`.
    network: Optional. Immutable. The name of the Compute Engine [VPC
      network](https://cloud.google.com/vpc/docs/vpc) to which the instance is
      connected.
    reservedIpRange: Optional. Immutable. The ID of the IP address range being
      used by the instance's VPC network. See [Configure a VPC network](https:
      //cloud.google.com/parallelstore/docs/vpc#create_and_configure_the_vpc).
      If no ID is provided, all ranges are considered.
    state: Output only. The instance state.
    updateTime: Output only. The time when the instance was updated.
  """

  class DeploymentTypeValueValuesEnum(_messages.Enum):
    r"""Optional. Immutable. The deployment type of the instance. Allowed
    values are: * `SCRATCH`: the instance is a scratch instance. *
    `PERSISTENT`: the instance is a persistent instance.

    Values:
      DEPLOYMENT_TYPE_UNSPECIFIED: Default Deployment Type It is equivalent to
        SCRATCH
      SCRATCH: Scratch
      PERSISTENT: Persistent
    """
    DEPLOYMENT_TYPE_UNSPECIFIED = 0
    SCRATCH = 1
    PERSISTENT = 2

  class DirectoryStripeLevelValueValuesEnum(_messages.Enum):
    r"""Optional. Immutable. Stripe level for directories. Allowed values are:
    * `DIRECTORY_STRIPE_LEVEL_MIN`: recommended when directories contain a
    small number of files. * `DIRECTORY_STRIPE_LEVEL_BALANCED`: balances
    performance for workloads involving a mix of small and large directories.
    * `DIRECTORY_STRIPE_LEVEL_MAX`: recommended for directories with a large
    number of files.

    Values:
      DIRECTORY_STRIPE_LEVEL_UNSPECIFIED: If not set, DirectoryStripeLevel
        will default to DIRECTORY_STRIPE_LEVEL_MAX
      DIRECTORY_STRIPE_LEVEL_MIN: Minimum directory striping
      DIRECTORY_STRIPE_LEVEL_BALANCED: Medium directory striping
      DIRECTORY_STRIPE_LEVEL_MAX: Maximum directory striping
    """
    DIRECTORY_STRIPE_LEVEL_UNSPECIFIED = 0
    DIRECTORY_STRIPE_LEVEL_MIN = 1
    DIRECTORY_STRIPE_LEVEL_BALANCED = 2
    DIRECTORY_STRIPE_LEVEL_MAX = 3

  class FileStripeLevelValueValuesEnum(_messages.Enum):
    r"""Optional. Immutable. Stripe level for files. Allowed values are: *
    `FILE_STRIPE_LEVEL_MIN`: offers the best performance for small size files.
    * `FILE_STRIPE_LEVEL_BALANCED`: balances performance for workloads
    involving a mix of small and large files. * `FILE_STRIPE_LEVEL_MAX`:
    higher throughput performance for larger files.

    Values:
      FILE_STRIPE_LEVEL_UNSPECIFIED: If not set, FileStripeLevel will default
        to FILE_STRIPE_LEVEL_BALANCED
      FILE_STRIPE_LEVEL_MIN: Minimum file striping
      FILE_STRIPE_LEVEL_BALANCED: Medium file striping
      FILE_STRIPE_LEVEL_MAX: Maximum file striping
    """
    FILE_STRIPE_LEVEL_UNSPECIFIED = 0
    FILE_STRIPE_LEVEL_MIN = 1
    FILE_STRIPE_LEVEL_BALANCED = 2
    FILE_STRIPE_LEVEL_MAX = 3

  class StateValueValuesEnum(_messages.Enum):
    r"""Output only. The instance state.

    Values:
      STATE_UNSPECIFIED: Not set.
      CREATING: The instance is being created.
      ACTIVE: The instance is available for use.
      DELETING: The instance is being deleted.
      FAILED: The instance is not usable.
      UPGRADING: The instance is being upgraded.
      REPAIRING: The instance is being repaired. This should only be used by
        instances using the `PERSISTENT` deployment type.
    """
    STATE_UNSPECIFIED = 0
    CREATING = 1
    ACTIVE = 2
    DELETING = 3
    FAILED = 4
    UPGRADING = 5
    REPAIRING = 6

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Optional. Cloud Labels are a flexible and lightweight mechanism for
    organizing cloud resources into groups that reflect a customer's
    organizational needs and deployment strategies. See
    https://cloud.google.com/resource-manager/docs/labels-overview for
    details.

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  accessPoints = _messages.StringField(1, repeated=True)
  capacityGib = _messages.IntegerField(2)
  createTime = _messages.StringField(3)
  daosVersion = _messages.StringField(4)
  deploymentType = _messages.EnumField('DeploymentTypeValueValuesEnum', 5)
  description = _messages.StringField(6)
  directoryStripeLevel = _messages.EnumField('DirectoryStripeLevelValueValuesEnum', 7)
  effectiveReservedIpRange = _messages.StringField(8)
  fileStripeLevel = _messages.EnumField('FileStripeLevelValueValuesEnum', 9)
  labels = _messages.MessageField('LabelsValue', 10)
  name = _messages.StringField(11)
  network = _messages.StringField(12)
  reservedIpRange = _messages.StringField(13)
  state = _messages.EnumField('StateValueValuesEnum', 14)
  updateTime = _messages.StringField(15)


class ListInstancesResponse(_messages.Message):
  r"""Response from ListInstances.

  Fields:
    instances: The list of Parallelstore instances.
    nextPageToken: A token identifying a page of results the server should
      return.
    unreachable: Locations that could not be reached.
  """

  instances = _messages.MessageField('Instance', 1, repeated=True)
  nextPageToken = _messages.StringField(2)
  unreachable = _messages.StringField(3, repeated=True)


class ListLocationsResponse(_messages.Message):
  r"""The response message for Locations.ListLocations.

  Fields:
    locations: A list of locations that matches the specified filter in the
      request.
    nextPageToken: The standard List next-page token.
  """

  locations = _messages.MessageField('Location', 1, repeated=True)
  nextPageToken = _messages.StringField(2)


class ListOperationsResponse(_messages.Message):
  r"""The response message for Operations.ListOperations.

  Fields:
    nextPageToken: The standard List next-page token.
    operations: A list of operations that matches the specified filter in the
      request.
    unreachable: Unordered list. Unreachable resources. Populated when the
      request sets `ListOperationsRequest.return_partial_success` and reads
      across collections. For example, when attempting to list all resources
      across all supported locations.
  """

  nextPageToken = _messages.StringField(1)
  operations = _messages.MessageField('Operation', 2, repeated=True)
  unreachable = _messages.StringField(3, repeated=True)


class Location(_messages.Message):
  r"""A resource that represents a Google Cloud location.

  Messages:
    LabelsValue: Cross-service attributes for the location. For example
      {"cloud.googleapis.com/region": "us-east1"}
    MetadataValue: Service-specific metadata. For example the available
      capacity at the given location.

  Fields:
    displayName: The friendly name for this location, typically a nearby city
      name. For example, "Tokyo".
    labels: Cross-service attributes for the location. For example
      {"cloud.googleapis.com/region": "us-east1"}
    locationId: The canonical id for this location. For example: `"us-east1"`.
    metadata: Service-specific metadata. For example the available capacity at
      the given location.
    name: Resource name for the location, which may vary between
      implementations. For example: `"projects/example-project/locations/us-
      east1"`
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class LabelsValue(_messages.Message):
    r"""Cross-service attributes for the location. For example
    {"cloud.googleapis.com/region": "us-east1"}

    Messages:
      AdditionalProperty: An additional property for a LabelsValue object.

    Fields:
      additionalProperties: Additional properties of type LabelsValue
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a LabelsValue object.

      Fields:
        key: Name of the additional property.
        value: A string attribute.
      """

      key = _messages.StringField(1)
      value = _messages.StringField(2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class MetadataValue(_messages.Message):
    r"""Service-specific metadata. For example the available capacity at the
    given location.

    Messages:
      AdditionalProperty: An additional property for a MetadataValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a MetadataValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  displayName = _messages.StringField(1)
  labels = _messages.MessageField('LabelsValue', 2)
  locationId = _messages.StringField(3)
  metadata = _messages.MessageField('MetadataValue', 4)
  name = _messages.StringField(5)


class Operation(_messages.Message):
  r"""This resource represents a long-running operation that is the result of
  a network API call.

  Messages:
    MetadataValue: Service-specific metadata associated with the operation. It
      typically contains progress information and common metadata such as
      create time. Some services might not provide such metadata. Any method
      that returns a long-running operation should document the metadata type,
      if any.
    ResponseValue: The normal, successful response of the operation. If the
      original method returns no data on success, such as `Delete`, the
      response is `google.protobuf.Empty`. If the original method is standard
      `Get`/`Create`/`Update`, the response should be the resource. For other
      methods, the response should have the type `XxxResponse`, where `Xxx` is
      the original method name. For example, if the original method name is
      `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.

  Fields:
    done: If the value is `false`, it means the operation is still in
      progress. If `true`, the operation is completed, and either `error` or
      `response` is available.
    error: The error result of the operation in case of failure or
      cancellation.
    metadata: Service-specific metadata associated with the operation. It
      typically contains progress information and common metadata such as
      create time. Some services might not provide such metadata. Any method
      that returns a long-running operation should document the metadata type,
      if any.
    name: The server-assigned name, which is only unique within the same
      service that originally returns it. If you use the default HTTP mapping,
      the `name` should be a resource name ending with
      `operations/{unique_id}`.
    response: The normal, successful response of the operation. If the
      original method returns no data on success, such as `Delete`, the
      response is `google.protobuf.Empty`. If the original method is standard
      `Get`/`Create`/`Update`, the response should be the resource. For other
      methods, the response should have the type `XxxResponse`, where `Xxx` is
      the original method name. For example, if the original method name is
      `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class MetadataValue(_messages.Message):
    r"""Service-specific metadata associated with the operation. It typically
    contains progress information and common metadata such as create time.
    Some services might not provide such metadata. Any method that returns a
    long-running operation should document the metadata type, if any.

    Messages:
      AdditionalProperty: An additional property for a MetadataValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a MetadataValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  @encoding.MapUnrecognizedFields('additionalProperties')
  class ResponseValue(_messages.Message):
    r"""The normal, successful response of the operation. If the original
    method returns no data on success, such as `Delete`, the response is
    `google.protobuf.Empty`. If the original method is standard
    `Get`/`Create`/`Update`, the response should be the resource. For other
    methods, the response should have the type `XxxResponse`, where `Xxx` is
    the original method name. For example, if the original method name is
    `TakeSnapshot()`, the inferred response type is `TakeSnapshotResponse`.

    Messages:
      AdditionalProperty: An additional property for a ResponseValue object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a ResponseValue object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  done = _messages.BooleanField(1)
  error = _messages.MessageField('Status', 2)
  metadata = _messages.MessageField('MetadataValue', 3)
  name = _messages.StringField(4)
  response = _messages.MessageField('ResponseValue', 5)


class OperationMetadata(_messages.Message):
  r"""Long-running operation metadata.

  Fields:
    apiVersion: Output only. API version used to start the operation.
    createTime: Output only. The time the operation was created.
    endTime: Output only. The time the operation finished running.
    requestedCancellation: Output only. Identifies whether the user has
      requested cancellation of the operation. Operations that have been
      cancelled successfully have Operation.error value with a
      google.rpc.Status.code of 1, corresponding to `Code.CANCELLED`.
    statusMessage: Output only. Human-readable status of the operation, if
      any.
    target: Output only. Server-defined resource path for the target of the
      operation.
    verb: Output only. Name of the verb executed by the operation.
  """

  apiVersion = _messages.StringField(1)
  createTime = _messages.StringField(2)
  endTime = _messages.StringField(3)
  requestedCancellation = _messages.BooleanField(4)
  statusMessage = _messages.StringField(5)
  target = _messages.StringField(6)
  verb = _messages.StringField(7)


class ParallelstoreProjectsLocationsGetRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsGetRequest object.

  Fields:
    name: Resource name for the location.
  """

  name = _messages.StringField(1, required=True)


class ParallelstoreProjectsLocationsInstancesCreateRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsInstancesCreateRequest object.

  Fields:
    instance: A Instance resource to be passed as the request body.
    instanceId: Required. The name of the Parallelstore instance. * Must
      contain only lowercase letters, numbers, and hyphens. * Must start with
      a letter. * Must be between 1-63 characters. * Must end with a number or
      a letter. * Must be unique within the customer project / location
    parent: Required. The instance's project and location, in the format
      `projects/{project}/locations/{location}`. Locations map to Google Cloud
      zones; for example, `us-west1-b`.
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes since the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
  """

  instance = _messages.MessageField('Instance', 1)
  instanceId = _messages.StringField(2)
  parent = _messages.StringField(3, required=True)
  requestId = _messages.StringField(4)


class ParallelstoreProjectsLocationsInstancesDeleteRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsInstancesDeleteRequest object.

  Fields:
    name: Required. Name of the resource
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes after the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
  """

  name = _messages.StringField(1, required=True)
  requestId = _messages.StringField(2)


class ParallelstoreProjectsLocationsInstancesExportDataRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsInstancesExportDataRequest object.

  Fields:
    exportDataRequest: A ExportDataRequest resource to be passed as the
      request body.
    name: Required. Name of the resource.
  """

  exportDataRequest = _messages.MessageField('ExportDataRequest', 1)
  name = _messages.StringField(2, required=True)


class ParallelstoreProjectsLocationsInstancesGetRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsInstancesGetRequest object.

  Fields:
    name: Required. The instance resource name, in the format
      `projects/{project_id}/locations/{location}/instances/{instance_id}`.
  """

  name = _messages.StringField(1, required=True)


class ParallelstoreProjectsLocationsInstancesImportDataRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsInstancesImportDataRequest object.

  Fields:
    importDataRequest: A ImportDataRequest resource to be passed as the
      request body.
    name: Required. Name of the resource.
  """

  importDataRequest = _messages.MessageField('ImportDataRequest', 1)
  name = _messages.StringField(2, required=True)


class ParallelstoreProjectsLocationsInstancesListRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsInstancesListRequest object.

  Fields:
    filter: Optional. Filtering results.
    orderBy: Optional. Hint for how to order the results.
    pageSize: Optional. Requested page size. Server may return fewer items
      than requested. If unspecified, the server will pick an appropriate
      default.
    pageToken: Optional. A token identifying a page of results the server
      should return.
    parent: Required. The project and location for which to retrieve instance
      information, in the format `projects/{project_id}/locations/{location}`.
      To retrieve instance information for all locations, use "-" as the value
      of `{location}`.
  """

  filter = _messages.StringField(1)
  orderBy = _messages.StringField(2)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  parent = _messages.StringField(5, required=True)


class ParallelstoreProjectsLocationsInstancesPatchRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsInstancesPatchRequest object.

  Fields:
    instance: A Instance resource to be passed as the request body.
    name: Identifier. The resource name of the instance, in the format
      `projects/{project}/locations/{location}/instances/{instance_id}`.
    requestId: Optional. An optional request ID to identify requests. Specify
      a unique request ID so that if you must retry your request, the server
      will know to ignore the request if it has already been completed. The
      server will guarantee that for at least 60 minutes since the first
      request. For example, consider a situation where you make an initial
      request and the request times out. If you make the request again with
      the same request ID, the server can check if original operation with the
      same request ID was received, and if so, will ignore the second request.
      This prevents clients from accidentally creating duplicate commitments.
      The request ID must be a valid UUID with the exception that zero UUID is
      not supported (00000000-0000-0000-0000-000000000000).
    updateMask: Required. Mask of fields to update. Field mask is used to
      specify the fields to be overwritten in the Instance resource by the
      update. At least one path must be supplied in this field. The fields
      specified in the update_mask are relative to the resource, not the full
      request.
  """

  instance = _messages.MessageField('Instance', 1)
  name = _messages.StringField(2, required=True)
  requestId = _messages.StringField(3)
  updateMask = _messages.StringField(4)


class ParallelstoreProjectsLocationsListRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsListRequest object.

  Fields:
    extraLocationTypes: Optional. Do not use this field unless explicitly
      documented otherwise. This is primarily for internal usage.
    filter: A filter to narrow down results to a preferred subset. The
      filtering language accepts strings like `"displayName=tokyo"`, and is
      documented in more detail in [AIP-160](https://google.aip.dev/160).
    name: The resource that owns the locations collection, if applicable.
    pageSize: The maximum number of results to return. If not set, the service
      selects a default.
    pageToken: A page token received from the `next_page_token` field in the
      response. Send that page token to receive the subsequent page.
  """

  extraLocationTypes = _messages.StringField(1, repeated=True)
  filter = _messages.StringField(2)
  name = _messages.StringField(3, required=True)
  pageSize = _messages.IntegerField(4, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(5)


class ParallelstoreProjectsLocationsOperationsCancelRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsOperationsCancelRequest object.

  Fields:
    cancelOperationRequest: A CancelOperationRequest resource to be passed as
      the request body.
    name: The name of the operation resource to be cancelled.
  """

  cancelOperationRequest = _messages.MessageField('CancelOperationRequest', 1)
  name = _messages.StringField(2, required=True)


class ParallelstoreProjectsLocationsOperationsDeleteRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsOperationsDeleteRequest object.

  Fields:
    name: The name of the operation resource to be deleted.
  """

  name = _messages.StringField(1, required=True)


class ParallelstoreProjectsLocationsOperationsGetRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsOperationsGetRequest object.

  Fields:
    name: The name of the operation resource.
  """

  name = _messages.StringField(1, required=True)


class ParallelstoreProjectsLocationsOperationsListRequest(_messages.Message):
  r"""A ParallelstoreProjectsLocationsOperationsListRequest object.

  Fields:
    filter: The standard list filter.
    name: The name of the operation's parent resource.
    pageSize: The standard list page size.
    pageToken: The standard list page token.
    returnPartialSuccess: When set to `true`, operations that are reachable
      are returned as normal, and those that are unreachable are returned in
      the ListOperationsResponse.unreachable field. This can only be `true`
      when reading across collections. For example, when `parent` is set to
      `"projects/example/locations/-"`. This field is not supported by default
      and will result in an `UNIMPLEMENTED` error if set unless explicitly
      documented otherwise in service or product specific documentation.
  """

  filter = _messages.StringField(1)
  name = _messages.StringField(2, required=True)
  pageSize = _messages.IntegerField(3, variant=_messages.Variant.INT32)
  pageToken = _messages.StringField(4)
  returnPartialSuccess = _messages.BooleanField(5)


class ReconciliationOperationMetadata(_messages.Message):
  r"""Operation metadata returned by the CLH during resource state
  reconciliation.

  Enums:
    ExclusiveActionValueValuesEnum: Excluisive action returned by the CLH.

  Fields:
    deleteResource: DEPRECATED. Use exclusive_action instead.
    exclusiveAction: Excluisive action returned by the CLH.
  """

  class ExclusiveActionValueValuesEnum(_messages.Enum):
    r"""Excluisive action returned by the CLH.

    Values:
      UNKNOWN_REPAIR_ACTION: Unknown repair action.
      DELETE: The resource has to be deleted. When using this bit, the CLH
        should fail the operation. DEPRECATED. Instead use DELETE_RESOURCE
        OperationSignal in SideChannel.
      RETRY: This resource could not be repaired but the repair should be
        tried again at a later time. This can happen if there is a dependency
        that needs to be resolved first- e.g. if a parent resource must be
        repaired before a child resource.
    """
    UNKNOWN_REPAIR_ACTION = 0
    DELETE = 1
    RETRY = 2

  deleteResource = _messages.BooleanField(1)
  exclusiveAction = _messages.EnumField('ExclusiveActionValueValuesEnum', 2)


class SourceGcsBucket(_messages.Message):
  r"""Cloud Storage as the source of a data transfer.

  Fields:
    uri: Required. URI to a Cloud Storage bucket in the format: `gs:///`. The
      path inside the bucket is optional.
  """

  uri = _messages.StringField(1)


class SourceParallelstore(_messages.Message):
  r"""Parallelstore as the source of a data transfer.

  Fields:
    path: Optional. Root directory path to the Paralellstore filesystem,
      starting with `/`. Defaults to `/` if unset.
  """

  path = _messages.StringField(1)


class StandardQueryParameters(_messages.Message):
  r"""Query parameters accepted by all methods.

  Enums:
    FXgafvValueValuesEnum: V1 error format.
    AltValueValuesEnum: Data format for response.

  Fields:
    f__xgafv: V1 error format.
    access_token: OAuth access token.
    alt: Data format for response.
    callback: JSONP
    fields: Selector specifying which fields to include in a partial response.
    key: API key. Your API key identifies your project and provides you with
      API access, quota, and reports. Required unless you provide an OAuth 2.0
      token.
    oauth_token: OAuth 2.0 token for the current user.
    prettyPrint: Returns response with indentations and line breaks.
    quotaUser: Available to use for quota purposes for server-side
      applications. Can be any arbitrary string assigned to a user, but should
      not exceed 40 characters.
    trace: A tracing token of the form "token:<tokenid>" to include in api
      requests.
    uploadType: Legacy upload protocol for media (e.g. "media", "multipart").
    upload_protocol: Upload protocol for media (e.g. "raw", "multipart").
  """

  class AltValueValuesEnum(_messages.Enum):
    r"""Data format for response.

    Values:
      json: Responses with Content-Type of application/json
      media: Media download with context-dependent Content-Type
      proto: Responses with Content-Type of application/x-protobuf
    """
    json = 0
    media = 1
    proto = 2

  class FXgafvValueValuesEnum(_messages.Enum):
    r"""V1 error format.

    Values:
      _1: v1 error format
      _2: v2 error format
    """
    _1 = 0
    _2 = 1

  f__xgafv = _messages.EnumField('FXgafvValueValuesEnum', 1)
  access_token = _messages.StringField(2)
  alt = _messages.EnumField('AltValueValuesEnum', 3, default='json')
  callback = _messages.StringField(4)
  fields = _messages.StringField(5)
  key = _messages.StringField(6)
  oauth_token = _messages.StringField(7)
  prettyPrint = _messages.BooleanField(8, default=True)
  quotaUser = _messages.StringField(9)
  trace = _messages.StringField(10)
  uploadType = _messages.StringField(11)
  upload_protocol = _messages.StringField(12)


class Status(_messages.Message):
  r"""The `Status` type defines a logical error model that is suitable for
  different programming environments, including REST APIs and RPC APIs. It is
  used by [gRPC](https://github.com/grpc). Each `Status` message contains
  three pieces of data: error code, error message, and error details. You can
  find out more about this error model and how to work with it in the [API
  Design Guide](https://cloud.google.com/apis/design/errors).

  Messages:
    DetailsValueListEntry: A DetailsValueListEntry object.

  Fields:
    code: The status code, which should be an enum value of google.rpc.Code.
    details: A list of messages that carry the error details. There is a
      common set of message types for APIs to use.
    message: A developer-facing error message, which should be in English. Any
      user-facing error message should be localized and sent in the
      google.rpc.Status.details field, or localized by the client.
  """

  @encoding.MapUnrecognizedFields('additionalProperties')
  class DetailsValueListEntry(_messages.Message):
    r"""A DetailsValueListEntry object.

    Messages:
      AdditionalProperty: An additional property for a DetailsValueListEntry
        object.

    Fields:
      additionalProperties: Properties of the object. Contains field @type
        with type URL.
    """

    class AdditionalProperty(_messages.Message):
      r"""An additional property for a DetailsValueListEntry object.

      Fields:
        key: Name of the additional property.
        value: A extra_types.JsonValue attribute.
      """

      key = _messages.StringField(1)
      value = _messages.MessageField('extra_types.JsonValue', 2)

    additionalProperties = _messages.MessageField('AdditionalProperty', 1, repeated=True)

  code = _messages.IntegerField(1, variant=_messages.Variant.INT32)
  details = _messages.MessageField('DetailsValueListEntry', 2, repeated=True)
  message = _messages.StringField(3)


class TransferMetadataOptions(_messages.Message):
  r"""Transfer metadata options for the instance.

  Enums:
    GidValueValuesEnum: Optional. The GID preservation behavior.
    ModeValueValuesEnum: Optional. The mode preservation behavior.
    UidValueValuesEnum: Optional. The UID preservation behavior.

  Fields:
    gid: Optional. The GID preservation behavior.
    mode: Optional. The mode preservation behavior.
    uid: Optional. The UID preservation behavior.
  """

  class GidValueValuesEnum(_messages.Enum):
    r"""Optional. The GID preservation behavior.

    Values:
      GID_UNSPECIFIED: default is GID_NUMBER_PRESERVE.
      GID_SKIP: Do not preserve GID during a transfer job.
      GID_NUMBER_PRESERVE: Preserve GID that is in number format during a
        transfer job.
    """
    GID_UNSPECIFIED = 0
    GID_SKIP = 1
    GID_NUMBER_PRESERVE = 2

  class ModeValueValuesEnum(_messages.Enum):
    r"""Optional. The mode preservation behavior.

    Values:
      MODE_UNSPECIFIED: default is MODE_PRESERVE.
      MODE_SKIP: Do not preserve mode during a transfer job.
      MODE_PRESERVE: Preserve mode during a transfer job.
    """
    MODE_UNSPECIFIED = 0
    MODE_SKIP = 1
    MODE_PRESERVE = 2

  class UidValueValuesEnum(_messages.Enum):
    r"""Optional. The UID preservation behavior.

    Values:
      UID_UNSPECIFIED: default is UID_NUMBER_PRESERVE.
      UID_SKIP: Do not preserve UID during a transfer job.
      UID_NUMBER_PRESERVE: Preserve UID that is in number format during a
        transfer job.
    """
    UID_UNSPECIFIED = 0
    UID_SKIP = 1
    UID_NUMBER_PRESERVE = 2

  gid = _messages.EnumField('GidValueValuesEnum', 1)
  mode = _messages.EnumField('ModeValueValuesEnum', 2)
  uid = _messages.EnumField('UidValueValuesEnum', 3)


encoding.AddCustomJsonFieldMapping(
    StandardQueryParameters, 'f__xgafv', '$.xgafv')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_1', '1')
encoding.AddCustomJsonEnumMapping(
    StandardQueryParameters.FXgafvValueValuesEnum, '_2', '2')



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for `gcloud network-connectivity`."""


import argparse
from typing import Any

from googlecloudsdk.core import exceptions
from googlecloudsdk.core import properties
import googlecloudsdk.generated_clients.apis.networkconnectivity.v1.networkconnectivity_v1_messages as v1
import googlecloudsdk.generated_clients.apis.networkconnectivity.v1beta.networkconnectivity_v1beta_messages as v1beta


# Constants
PROJECTS_RESOURCE_PATH = "projects/"
LOCATION_FILTER_FMT = "location:projects/{0}/locations/{1}"
ROUTE_TYPE_FILTER = "-type:DYNAMIC_ROUTE"
LOCATION_TEMPLATE = "projects/{}/locations/{}"


class NetworkConnectivityError(exceptions.Error):
  """Top-level exception for all Network Connectivity errors."""


class InvalidInputError(NetworkConnectivityError):
  """Exception for invalid input."""


# Table format for spokes list
LIST_FORMAT = """
    table(
      name.basename(),
      name.segment(3):label=LOCATION,
      hub.basename(),
      group.basename(),
      format(
        "{0}{1}{2}{3}{4}",
        linkedVpnTunnels.yesno(yes="VPN tunnel", no=""),
        linkedInterconnectAttachments.yesno(yes="VLAN attachment", no=""),
        linkedRouterApplianceInstances.yesno(yes="Router appliance", no=""),
        linkedVpcNetwork.yesno(yes="VPC network", no=""),
        gateway.yesno(yes="Gateway", no="")
      ):label=TYPE,
      firstof(linkedVpnTunnels.uris, linkedInterconnectAttachments.uris, linkedRouterApplianceInstances.instances).len().yesno(no="1"):label="RESOURCE COUNT",
      format(
        "{0}{1}",
        linkedVpcNetwork.yesno(yes="N/A", no=""),
        firstof(linkedVpnTunnels.siteToSiteDataTransfer, linkedInterconnectAttachments.siteToSiteDataTransfer, linkedRouterApplianceInstances.siteToSiteDataTransfer, Gateway).yesno(yes="On", no="")
      ).yesno(no="Off"):label="DATA TRANSFER",
      description
    )
"""

LIST_SPOKES_FORMAT = """
    table(
      name.basename(),
      group.basename(),
      name.segment(1):label=PROJECT,
      name.segment(3):label=LOCATION,
      spokeType:label=TYPE,
      state,
      reasons.code.list():label="STATE REASON",
      etag,
      format(
        "{0}{1}",
        linkedVpcNetwork.yesno(yes="N/A", no=""),
        firstof(linkedVpnTunnels.siteToSiteDataTransfer, linkedInterconnectAttachments.siteToSiteDataTransfer, linkedRouterApplianceInstances.siteToSiteDataTransfer, gateway).yesno(yes="On", no="")
      ).yesno(no="Off").if(view=detailed):label="DATA TRANSFER",
      description.if(view=detailed)
    )
"""

LIST_TRANSPORTS_FORMAT = """
    table(
      name.basename(),
      remoteProfile.basename():label=PROFILE,
      bandwidth:label=BANDWIDTH,
      state:label=STATE
    )
"""

LIST_REMOTE_PROFILES_FORMAT = """
    table(
      name.basename(),
      provider:label=PROVIDER,
      providerSite:label=LOCATION
    )
"""


def AppendLocationsGlobalToParent(unused_ref, unused_args, request):
  """Add locations/global to parent path."""

  request.parent += "/locations/global"
  return request


def SetExportPscBeta(unused_ref, args, request):
  """Set legacy export_psc field based on new PSC flags."""
  if not args.IsSpecified("export_psc"):
    # If either of the new flags are specified, set export_psc to true too.
    if (
        args.export_psc_published_services_and_regional_google_apis
        or args.export_psc_global_google_apis
    ):
      request.googleCloudNetworkconnectivityV1betaHub.exportPsc = True
  return request


def DeriveProjectFromResource(resource):
  """Returns the project from a resource string."""
  if PROJECTS_RESOURCE_PATH not in resource:
    raise InvalidInputError(
        "Resource must contain a project path, but received: {0}".format(
            resource
        )
    )
  project = resource[
      resource.index(PROJECTS_RESOURCE_PATH) + len(PROJECTS_RESOURCE_PATH) :
  ]
  project = project.split("/")[0]
  return project


def AppendEffectiveLocationFilter(unused_ref, args, request):
  """Append filter to limit listing dynamic routes at an effective location."""

  if args.IsSpecified("effective_location"):
    location = args.effective_location
    project = DeriveProjectFromResource(request.parent)
    location_filter = LOCATION_FILTER_FMT.format(project, location)
    request.filter = "{0} OR {1}".format(location_filter, ROUTE_TYPE_FILTER)
  return request


def SetGlobalLocation():
  """Set default location to global."""
  return "global"


def ClearOverlaps(unused_ref, args, patch_request):
  """Handles clear_overlaps flag."""

  if args.IsSpecified("clear_overlaps"):
    if patch_request.updateMask:
      patch_request.updateMask += ",overlaps"
    else:
      patch_request.updateMask = "overlaps"
  return patch_request


def ClearLabels(unused_ref, args, patch_request):
  """Handles clear_labels flag."""

  if args.IsSpecified("clear_labels"):
    if patch_request.updateMask:
      patch_request.updateMask += ",labels"
    else:
      patch_request.updateMask = "labels"
  return patch_request


def ValidateMigration(unused_ref, args, request):
  """Validates internal range migration parameters."""
  if not args.IsSpecified("usage") or args.usage != "for-migration":
    if args.IsSpecified("migration_source") or args.IsSpecified(
        "migration_target"
    ):
      raise InvalidInputError(
          "migration_source and migration_target can only be specified when"
          " usage is set to for-migration."
      )
    else:
      return request
  # We are now in the for-migration usage case.
  if not args.IsSpecified("migration_source") or not args.IsSpecified(
      "migration_target"
  ):
    raise InvalidInputError(
        "Both migration_source and migration_target must be specified."
    )
  if args.IsSpecified("peering") and args.peering != "for-self":
    raise InvalidInputError(
        "Peering must be set to for-self when usage is set to for-migration."
    )
  if not args.migration_source:
    raise InvalidInputError("migration_source cannot be empty.")
  if not args.migration_target:
    raise InvalidInputError("migration_target cannot be empty.")
  if args.migration_source == args.migration_target:
    raise InvalidInputError(
        "migration_source and migration_target cannot be the same."
    )
  return request


def ValidateAllocationOptions(ref: Any, args: argparse.Namespace, request: Any):
  """Validates internal range allocation options."""

  del ref  # Unused.
  if (
      args.IsSpecified("allocation_strategy")
      and args.allocation_strategy == "RANDOM_FIRST_N_AVAILABLE"
  ):
    if (
        not args.IsSpecified("first_available_ranges_lookup_size")
        or args.first_available_ranges_lookup_size < 1
    ):
      raise InvalidInputError(
          "first_available_ranges_lookup_size must be set and greater than 0"
          " when allocation_strategy is RANDOM_FIRST_N_AVAILABLE."
      )
  elif args.IsSpecified("first_available_ranges_lookup_size"):
    raise InvalidInputError(
        "first_available_ranges_lookup_size can only be set when"
        " allocation_strategy is RANDOM_FIRST_N_AVAILABLE."
    )
  return request


class StoreGlobalAction(argparse._StoreConstAction):
  # pylint: disable=protected-access
  # pylint: disable=redefined-builtin
  """Return "global" if the --global argument is used."""

  def __init__(
      self, option_strings, dest, default="", required=False, help=None
  ):
    super(StoreGlobalAction, self).__init__(
        option_strings=option_strings,
        dest=dest,
        const="global",
        default=default,
        required=required,
        help=help,
    )


def SetGatewayAdvertisedRouteRecipient(unused_ref, args, request):
  """Set the route's `recipient` field based on boolean flags.

  Args:
    args: The command arguments.
    request: The request to set the `recipient` field on.

  Returns:
    The request with the `recipient` field set.
  """
  if args.advertise_to_hub:
    if (
        hasattr(request, "gatewayAdvertisedRoute")
        and request.gatewayAdvertisedRoute
    ):
      request.gatewayAdvertisedRoute.recipient = (
          v1.GatewayAdvertisedRoute.RecipientValueValuesEnum.ADVERTISE_TO_HUB
      )
    elif (
        hasattr(
            request,
            "googleCloudNetworkconnectivityV1betaGatewayAdvertisedRoute",
        )
        and request.googleCloudNetworkconnectivityV1betaGatewayAdvertisedRoute
    ):
      route = request.googleCloudNetworkconnectivityV1betaGatewayAdvertisedRoute
      enum_cls = (
          v1beta.GoogleCloudNetworkconnectivityV1betaGatewayAdvertisedRoute
          .RecipientValueValuesEnum
      )
      route.recipient = enum_cls.ADVERTISE_TO_HUB
  return request


def CheckRegionSpecifiedIfSpokeSpecified(unused_ref, unused_args, request):
  """If a spoke name is specified, then its region must also be specified.

  This is because CCFE doesn't support a wildcard ("-") in this case but returns
  a confusing error message. So we give the user a friendlier error.

  Args:
    request: The request object. We will inspect the parent field.

  Returns:
    The unmodified request object.
  Raises:
    InvalidInputError: If the region is unspecified when a spoke is.
  """
  region_wildcard = "/locations/-/" in request.parent
  spoke_wildcard = request.parent.endswith("/spokes/-")
  if region_wildcard and not spoke_wildcard:
    raise InvalidInputError(
        "A region must be specified if a spoke name is specified"
    )
  return request


def CheckForRouteTableAndHubWildcardMismatch(unused_ref, unused_args, request):
  """Check that hub and route table are both specified or both unspecified.

  This is because CCFE doesn't support wildcards ("-") in this case but returns
  a confusing error message. So we give he user a friendlier error.

  Args:
   request: The request object.

  Returns:
    The unmodified request object.
  Raises:
    InvalidInputError: If the user needs to specify a hub name or route table
    name.
  """
  hub_wildcard = "/hubs/-/" in request.parent
  route_table_wildcard = request.parent.endswith("/routeTables/-")
  if hub_wildcard and not route_table_wildcard:
    raise InvalidInputError(
        "A hub must be specified if a route table is specified"
    )
  if route_table_wildcard and not hub_wildcard:
    raise InvalidInputError(
        "A route table must be specified if a hub is specified"
    )
  return request


def AddDefaultLocationToList(ref, args, req):
  """Python hook for yaml commands to wildcard the region in list requests."""
  del ref
  project = properties.VALUES.core.project.Get(required=True)

  req.parent = LOCATION_TEMPLATE.format(project, args.region)
  return req



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Resource definitions for Cloud Platform Apis generated from apitools."""

import enum


BASE_URL = 'https://networkservices.googleapis.com/v1alpha1/'
DOCS_URL = 'https://cloud.google.com/networking'


class Collections(enum.Enum):
  """Collections for all supported apis."""

  PROJECTS = (
      'projects',
      'projects/{projectsId}',
      {},
      ['projectsId'],
      True
  )
  PROJECTS_LOCATIONS = (
      'projects.locations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_AGENTGATEWAYS = (
      'projects.locations.agentGateways',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/agentGateways/'
              '{agentGatewaysId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_AUTHZEXTENSIONS = (
      'projects.locations.authzExtensions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/authzExtensions/'
              '{authzExtensionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_EDGECACHEKEYSETS = (
      'projects.locations.edgeCacheKeysets',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'edgeCacheKeysets/{edgeCacheKeysetsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_EDGECACHEORIGINS = (
      'projects.locations.edgeCacheOrigins',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'edgeCacheOrigins/{edgeCacheOriginsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_EDGECACHESERVICES = (
      'projects.locations.edgeCacheServices',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'edgeCacheServices/{edgeCacheServicesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_ENDPOINTPOLICIES = (
      'projects.locations.endpointPolicies',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'endpointPolicies/{endpointPoliciesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_EXPRESSLINKS = (
      'projects.locations.expressLinks',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/expressLinks/'
              '{expressLinksId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_EXTENSIONBINDINGS = (
      'projects.locations.extensionBindings',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'extensionBindings/{extensionBindingsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_GATEWAYS = (
      'projects.locations.gateways',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/gateways/'
              '{gatewaysId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_GATEWAYS_ROUTEVIEWS = (
      'projects.locations.gateways.routeViews',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/gateways/'
              '{gatewaysId}/routeViews/{routeViewsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_GLOBAL_GOOGLETAGGATEWAYPOLICIES = (
      'projects.locations.global.googleTagGatewayPolicies',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/global/'
              'googleTagGatewayPolicies/{googleTagGatewayPoliciesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_GRPCROUTES = (
      'projects.locations.grpcRoutes',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/grpcRoutes/'
              '{grpcRoutesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_HTTPFILTERS = (
      'projects.locations.httpFilters',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/httpFilters/'
              '{httpFiltersId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_HTTPROUTES = (
      'projects.locations.httpRoutes',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/httpRoutes/'
              '{httpRoutesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_LBEDGEEXTENSIONS = (
      'projects.locations.lbEdgeExtensions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'lbEdgeExtensions/{lbEdgeExtensionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_LBOBSERVABILITYEXTENSIONS = (
      'projects.locations.lbObservabilityExtensions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'lbObservabilityExtensions/{lbObservabilityExtensionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_LBROUTEEXTENSIONS = (
      'projects.locations.lbRouteExtensions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'lbRouteExtensions/{lbRouteExtensionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_LBTCPEXTENSIONS = (
      'projects.locations.lbTcpExtensions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/lbTcpExtensions/'
              '{lbTcpExtensionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_LBTRAFFICEXTENSIONS = (
      'projects.locations.lbTrafficExtensions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'lbTrafficExtensions/{lbTrafficExtensionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MESHES = (
      'projects.locations.meshes',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/meshes/'
              '{meshesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MESHES_ROUTEVIEWS = (
      'projects.locations.meshes.routeViews',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/meshes/'
              '{meshesId}/routeViews/{routeViewsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTCONSUMERASSOCIATIONS = (
      'projects.locations.multicastConsumerAssociations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastConsumerAssociations/'
              '{multicastConsumerAssociationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTDOMAINACTIVATIONS = (
      'projects.locations.multicastDomainActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastDomainActivations/{multicastDomainActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTDOMAINGROUPS = (
      'projects.locations.multicastDomainGroups',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastDomainGroups/{multicastDomainGroupsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTDOMAINS = (
      'projects.locations.multicastDomains',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastDomains/{multicastDomainsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTGROUPCONSUMERACTIVATIONS = (
      'projects.locations.multicastGroupConsumerActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastGroupConsumerActivations/'
              '{multicastGroupConsumerActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTGROUPDEFINITIONS = (
      'projects.locations.multicastGroupDefinitions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastGroupDefinitions/{multicastGroupDefinitionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTGROUPPRODUCERACTIVATIONS = (
      'projects.locations.multicastGroupProducerActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastGroupProducerActivations/'
              '{multicastGroupProducerActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTGROUPRANGEACTIVATIONS = (
      'projects.locations.multicastGroupRangeActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastGroupRangeActivations/'
              '{multicastGroupRangeActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTGROUPRANGES = (
      'projects.locations.multicastGroupRanges',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastGroupRanges/{multicastGroupRangesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTGROUPS = (
      'projects.locations.multicastGroups',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/multicastGroups/'
              '{multicastGroupsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_MULTICASTPRODUCERASSOCIATIONS = (
      'projects.locations.multicastProducerAssociations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'multicastProducerAssociations/'
              '{multicastProducerAssociationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_OBSERVABILITYPOLICIES = (
      'projects.locations.observabilityPolicies',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'observabilityPolicies/{observabilityPoliciesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_OPERATIONS = (
      'projects.locations.operations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/operations/'
              '{operationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_REGIONALMULTICASTCONSUMERASSOCIATIONS = (
      'projects.locations.regionalMulticastConsumerAssociations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'regionalMulticastConsumerAssociations/'
              '{regionalMulticastConsumerAssociationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_REGIONALMULTICASTDOMAINACTIVATIONS = (
      'projects.locations.regionalMulticastDomainActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'regionalMulticastDomainActivations/'
              '{regionalMulticastDomainActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_REGIONALMULTICASTGROUPCONSUMERACTIVATIONS = (
      'projects.locations.regionalMulticastGroupConsumerActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'regionalMulticastGroupConsumerActivations/'
              '{regionalMulticastGroupConsumerActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_REGIONALMULTICASTGROUPPRODUCERACTIVATIONS = (
      'projects.locations.regionalMulticastGroupProducerActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'regionalMulticastGroupProducerActivations/'
              '{regionalMulticastGroupProducerActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_REGIONALMULTICASTGROUPRANGEACTIVATIONS = (
      'projects.locations.regionalMulticastGroupRangeActivations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'regionalMulticastGroupRangeActivations/'
              '{regionalMulticastGroupRangeActivationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_REGIONALMULTICASTPRODUCERASSOCIATIONS = (
      'projects.locations.regionalMulticastProducerAssociations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'regionalMulticastProducerAssociations/'
              '{regionalMulticastProducerAssociationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_SERVICEBINDINGS = (
      'projects.locations.serviceBindings',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/serviceBindings/'
              '{serviceBindingsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_SERVICELBPOLICIES = (
      'projects.locations.serviceLbPolicies',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'serviceLbPolicies/{serviceLbPoliciesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_SWPSECURITYEXTENSIONS = (
      'projects.locations.swpSecurityExtensions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'swpSecurityExtensions/{swpSecurityExtensionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_TCPROUTES = (
      'projects.locations.tcpRoutes',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/tcpRoutes/'
              '{tcpRoutesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_TELEMETRYPOLICIES = (
      'projects.locations.telemetryPolicies',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'telemetryPolicies/{telemetryPoliciesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_TLSROUTES = (
      'projects.locations.tlsRoutes',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/tlsRoutes/'
              '{tlsRoutesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_WASMACTIONS = (
      'projects.locations.wasmActions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/wasmActions/'
              '{wasmActionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_WASMPLUGINS = (
      'projects.locations.wasmPlugins',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/wasmPlugins/'
              '{wasmPluginsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_WASMPLUGINS_VERSIONS = (
      'projects.locations.wasmPlugins.versions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/wasmPlugins/'
              '{wasmPluginsId}/versions/{versionsId}',
      },
      ['name'],
      True
  )

  def __init__(self, collection_name, path, flat_paths, params,
               enable_uri_parsing):
    self.collection_name = collection_name
    self.path = path
    self.flat_paths = flat_paths
    self.params = params
    self.enable_uri_parsing = enable_uri_parsing



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to describe a Cloud FTP user."""

from googlecloudsdk.api_lib.storage import ftp
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.storage.ftp import users_util


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Describe(base.DescribeCommand):
  """Describe an existing Cloud FTP user."""

  hints = base.CommandHint(read_only=True)

  detailed_help = {
      'DESCRIPTION': (
          """\
          Describe an existing Cloud FTP user.
      """
      ),
      'EXAMPLES': (
          """\
          To describe user `my-user` for server `my-server` in `us-west1`, run:

            $ {command} my-user --server=my-server --location=us-west1
      """
      ),
  }

  @staticmethod
  def Args(parser):
    parser.add_argument(
        'USER_ID',
        help='The ID of the user to describe.',
    )
    parser.add_argument(
        '--server',
        required=True,
        help='The ID of the server.',
    )
    parser.add_argument(
        '--location',
        required=True,
        help='The region of the server.',
    )

  def Run(self, args):
    client = ftp.FtpClient()
    user_name = users_util.GetUserResourceName(
        args.location, args.server, args.USER_ID
    )
    return client.GetUser(user_name)



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""API client library for Application Templates."""


from googlecloudsdk.api_lib.design_center import utils
from googlecloudsdk.calliope import base


class ApplicationTemplatesClient(object):
  """Client for Application Templates in the Design Center API."""

  def __init__(self, release_track=base.ReleaseTrack.ALPHA):
    self.client = utils.GetClientInstance(release_track)
    self.messages = utils.GetMessagesModule(release_track)
    self._service = self.client.projects_locations_spaces_applicationTemplates

  def ImportIac(self, name, gcs_uri=None, iac_module=None,
                allow_partial_import=False, validate_iac=False):
    """Calls the ImportApplicationTemplateIaC RPC.

    Args:
      name: str, The full resource name of the Application Template.
      gcs_uri: str, The GCS URI of the IaC source.
      iac_module: messages.IaCModule, The IaCModule object.
      allow_partial_import: bool, Whether to allow partial imports.
      validate_iac: bool, Whether to only validate the IaC.

    Returns:
      The response from the API call.
    """
    if not name:
      raise ValueError('Application template name cannot be empty or None.')

    import_iac_request = self.messages.ImportApplicationTemplateIaCRequest(
        allowPartialImport=allow_partial_import,
        validateIac=validate_iac)

    if gcs_uri:
      import_iac_request.gcsUri = gcs_uri
    elif iac_module:
      import_iac_request.iacModule = iac_module

    request = self.messages.DesigncenterProjectsLocationsSpacesApplicationTemplatesImportIaCRequest(
        name=name,
        importApplicationTemplateIaCRequest=import_iac_request)

    return self._service.ImportIaC(request)

  def GenerateAssessmentReport(self, name, *,
                               service_account=None):
    """Calls the GenerateApplicationTemplateAssessmentReport RPC.

    Args:
      name: The full resource name of the Application Template.
      service_account: The service account to use for the preview.

    Returns:
      The response from the API call.
    """
    return self._service.GenerateAssessmentReport(
        self.messages.DesigncenterProjectsLocationsSpacesApplicationTemplatesGenerateAssessmentReportRequest(
            name=name,
            generateApplicationTemplateAssessmentReportRequest=self.messages.GenerateApplicationTemplateAssessmentReportRequest(
                serviceAccount=service_account)))



# -*- coding: utf-8 -*- #
# Copyright 2024 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for updating multi-region Services."""

from googlecloudsdk.api_lib.run import k8s_object
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as c_exceptions
from googlecloudsdk.command_lib.run import config_changes
from googlecloudsdk.command_lib.run import connection_context
from googlecloudsdk.command_lib.run import flags
from googlecloudsdk.command_lib.run import platforms
from googlecloudsdk.command_lib.run import pretty_print
from surface.run.services import replace


@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
class MultiRegionReplace(replace.Replace):
  """Create or Update multi-region service from YAML."""

  @classmethod
  def Args(cls, parser):
    super().Args(parser)
    flags.AddRegionsArg(parser)
    flags.AddAddRegionsArg(parser)
    flags.AddRemoveRegionsArg(parser)

  def _GetMultiRegionSettings(self, args):
    added_or_removed = flags.FlagIsExplicitlySet(
        args, 'add_regions'
    ) or flags.FlagIsExplicitlySet(args, 'remove_regions')
    all_regions = flags.FlagIsExplicitlySet(args, 'regions')
    if added_or_removed and all_regions:
      raise c_exceptions.InvalidArgumentException(
          parameter_name='--regions',
          message=(
              'Cannot specify --add-regions or --remove-regions with --regions'
          ),
      )
    return all_regions, added_or_removed

  def _GetBaseChanges(self, new_service, args):
    if not new_service:
      return None
    all_regions, added_or_removed = self._GetMultiRegionSettings(args)
    changes = super()._GetBaseChanges(new_service, args)
    if added_or_removed:
      changes.append(
          config_changes.RegionsChangeAnnotationChange(
              to_add=args.add_regions,
              to_remove=args.remove_regions,
          )
      )
    if all_regions:
      changes.append(config_changes.SetRegionsAnnotationChange(args.regions))
    return changes

  def _ConnectionContext(self, args, region_label):
    return connection_context.GetConnectionContext(
        args,
        flags.Product.RUN,
        self.ReleaseTrack(),
        region_label=region_label,
        is_multiregion=True,
    )

  def _GetMultiRegionRegions(self, args, new_service, changes):
    if not new_service:
      return None
    # If the user set --regions, then we just use that. Otherwise, we need to
    # parse the changes to figure out what regions are being added/removed.
    all_regions, added_or_removed = self._GetMultiRegionSettings(args)
    if all_regions:
      return args.regions.split(',')
    if added_or_removed:
      modified = config_changes.WithChanges(new_service, changes)
      annotation = (
          modified.annotations.get(k8s_object.MULTI_REGION_REGIONS_ANNOTATION)
          or None
      )
      return annotation.split(',') if annotation else None
    return None

  def _PrintSuccessMessage(self, service_obj, dry_run, args):
    if args.async_:
      pretty_print.Success(
          'New configuration for [{{bold}}{serv}{{reset}}] is being applied '
          'asynchronously.'.format(serv=service_obj.name)
      )
    elif dry_run:
      pretty_print.Success(
          'New configuration has been validated for Multi-region service '
          '[{{bold}}{serv}{{reset}}].'.format(serv=service_obj.name)
      )
    else:
      pretty_print.Success(
          'New configuration has been applied to Multi-region service '
          '[{{bold}}{serv}{{reset}}].\nRegional URLs:'.format(
              serv=service_obj.name
          )
      )
      for url in service_obj.urls:
        pretty_print.Success('{{bold}}{url}{{reset}}'.format(url=url))

  def Run(self, args):
    if platforms.GetPlatform() != platforms.PLATFORM_MANAGED:
      raise c_exceptions.InvalidArgumentException(
          '--platform',
          'Multi-region Services are only supported on managed platform.',
      )
    if flags.FlagIsExplicitlySet(args, 'region'):
      raise c_exceptions.InvalidArgumentException(
          '--region',
          'Multi-region Services do not support the --region flag. Use'
          ' --regions, --add-regions, or --remove-regions instead.',
      )
    return super().Run(args)



# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tunnel TCP traffic over Cloud IAP WebSocket connection."""


import ctypes
import errno
import functools
import gc
import io
import os
import select
import signal
import socket
import sys
import threading

from googlecloudsdk.api_lib.compute import iap_tunnel_websocket
from googlecloudsdk.api_lib.compute import iap_tunnel_websocket_utils as utils
from googlecloudsdk.api_lib.compute import sg_tunnel
from googlecloudsdk.api_lib.compute import sg_tunnel_utils as sg_utils
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import execution_utils
from googlecloudsdk.core import http_proxy
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core import transport
from googlecloudsdk.core.credentials import creds
from googlecloudsdk.core.credentials import store
from googlecloudsdk.core.util import files
from googlecloudsdk.core.util import platforms
import portpicker
import six
from six.moves import queue


if not platforms.OperatingSystem.IsWindows():
  import fcntl  # pylint: disable=g-import-not-at-top
else:
  from ctypes import wintypes  # pylint: disable=g-import-not-at-top

ERROR_BROKEN_PIPE = 109
READ_FROM_STDIN_TIMEOUT_SECS = 3
_TERMINATION_SIGNALS = (
    'SIGHUP',
    'SIGTERM',
    'SIGINT',
    'SIGQUIT',
    'SIGBREAK',
    'SIGABRT',
)


class LocalPortUnavailableError(exceptions.Error):
  pass


class UnableToOpenPortError(exceptions.Error):
  pass


def _AddBaseArgs(parser):
  parser.add_argument(
      '--iap-tunnel-url-override',
      hidden=True,
      help=('Allows for overriding the connection endpoint for integration '
            'testing.'))
  parser.add_argument(
      '--iap-tunnel-insecure-disable-websocket-cert-check',
      default=False,
      action='store_true',
      hidden=True,
      help='Disables checking certificates on the WebSocket connection.')


def AddSshTunnelArgs(parser, tunnel_through_iap_scope):
  _AddBaseArgs(parser)
  tunnel_through_iap_scope.add_argument(
      '--tunnel-through-iap',
      action='store_true',
      help="""\
      Tunnel the ssh connection through Cloud Identity-Aware Proxy for TCP
      forwarding.

      To learn more, see the
      [IAP for TCP forwarding documentation](https://cloud.google.com/iap/docs/tcp-forwarding-overview).
      """)


def AddHostBasedTunnelArgs(parser, support_security_gateway=False):
  """Add the arguments for supporting host-based connections."""

  group = parser.add_argument_group()
  group.add_argument(
      '--region',
      default=None,
      required=True,
      help=('Configures the region to use when connecting via IP address or '
            'FQDN.'))

  if support_security_gateway:
    group_mutex = group.add_argument_group(mutex=True)
    AddSecurityGatewayTunnelArgs(group_mutex.add_argument_group(hidden=True))
    group = group_mutex.add_argument_group()
  AddOnPremTunnelArgs(group)


def AddOnPremTunnelArgs(parser):
  """Add the arguments for supporting IP/FQDN-based tunnels."""
  parser.add_argument(
      '--network',
      default=None,
      required=True,
      help=(
          'Configures the VPC network to use when connecting via IP address or '
          'FQDN.'))
  # TODO(b/196572980): Make dest-group required in beta/GA.
  parser.add_argument(
      '--dest-group',
      default=None,
      required=False,
      help=('Configures the destination group to use when connecting via IP '
            'address or FQDN.'))


def AddSecurityGatewayTunnelArgs(parser):
  """Add arguments for the Security Gateway path."""
  parser.add_argument(
      '--security-gateway',
      default=None,
      required=True,
      help='Configure the security gateway resource for connecting.')
  # TODO(b/196572980): Make dest-group required in beta/GA.
  parser.add_argument(
      '--use-dest-group',
      default=False,
      action='store_true',
      required=False,
      help=('Configures the destination group to use when connecting via IP '
            'address or FQDN.'))


def AddProxyServerHelperArgs(parser):
  _AddBaseArgs(parser)


def CreateSshTunnelArgs(args, track, instance_ref, external_interface):
  """Construct an SshTunnelArgs from command line args and values.

  Args:
    args: The parsed commandline arguments. May or may not have had
      AddSshTunnelArgs called.
    track: ReleaseTrack, The currently running release track.
    instance_ref: The target instance reference object.
    external_interface: The external interface of target resource object, if
      available, otherwise None.
  Returns:
    SshTunnelArgs or None if IAP Tunnel is disabled.
  """
  # If tunneling through IAP is not available, then abort.
  if not hasattr(args, 'tunnel_through_iap'):
    return None

  # If set to connect directly to private IP address, then abort.
  if getattr(args, 'internal_ip', False):
    return None

  if args.IsSpecified('tunnel_through_iap'):
    # If IAP tunneling is explicitly disabled, then abort.
    if not args.tunnel_through_iap:
      return None
  else:
    # If no external interface is available, then default to using IAP
    # tunneling and continue with code below.  Otherwise, abort.
    if external_interface:
      return None
    log.status.Print('External IP address was not found; defaulting to using '
                     'IAP tunneling.')

  res = SshTunnelArgs()

  res.track = track.prefix
  res.project = instance_ref.project
  res.zone = instance_ref.zone
  res.instance = instance_ref.instance

  _AddPassThroughArgs(args, res)

  return res


def CreateOnPremSshTunnelArgs(args, track, host):
  """Construct an SshTunnelArgs from command line args and values for on-prem.

  Args:
    args: The parsed commandline arguments. May or may not have had
      AddSshTunnelArgs called.
    track: ReleaseTrack, The currently running release track.
    host: The target IP address or FQDN.
  Returns:
    SshTunnelArgs.
  """
  res = SshTunnelArgs()

  res.track = track.prefix
  res.project = properties.VALUES.core.project.GetOrFail()
  res.region = args.region
  res.network = args.network
  res.instance = host

  _AddPassThroughArgs(args, res)

  return res


def _AddPassThroughArgs(args, ssh_tunnel_args):
  """Adds any passthrough args to the SshTunnelArgs.

  Args:
    args: The parsed commandline arguments. May or may not have had
      AddSshTunnelArgs called.
    ssh_tunnel_args: SshTunnelArgs, The SSH tunnel args to update.
  """
  if args.IsSpecified('iap_tunnel_url_override'):
    ssh_tunnel_args.pass_through_args.append(
        '--iap-tunnel-url-override=' + args.iap_tunnel_url_override)
  if args.iap_tunnel_insecure_disable_websocket_cert_check:
    ssh_tunnel_args.pass_through_args.append(
        '--iap-tunnel-insecure-disable-websocket-cert-check')
  if args.IsKnownAndSpecified('dest_group'):
    ssh_tunnel_args.pass_through_args.append('--dest-group=' + args.dest_group)


class SshTunnelArgs(object):
  """A class to hold some options for IAP Tunnel SSH/SCP.

  Attributes:
    track: str/None, the prefix of the track for the inner gcloud.
    project: str, the project id (string with dashes).
    zone: str, the zone name.
    instance: str, the instance name (or IP or FQDN for on-prem).
    region: str, the region name (on-prem only).
    network: str, the network name (on-prem only).
    cloud_run_args: dict, The fields required to construct Cloud Run
      SshTunnelArgs. If present, this field should contain fields for
      'deployment_name', 'workload_type', and 'project_number'. Optionally can
      contain 'instance_id' and 'container_id'.
    pass_through_args: [str], additional args to be passed to the inner gcloud.
  """

  def __init__(self):
    self.track = None
    self.project = ''
    self.zone = ''
    self.instance = ''
    self.region = ''
    self.network = ''
    self.cloud_run_args = None
    self.pass_through_args = []

  def _Members(self):
    return (
        self.track,
        self.project,
        self.zone,
        self.instance,
        self.region,
        self.network,
        self.pass_through_args,
        self.cloud_run_args,
    )

  def __eq__(self, other):
    # pylint: disable=protected-access
    return self._Members() == other._Members()

  def __ne__(self, other):
    return not self == other

  def __repr__(self):
    return 'SshTunnelArgs<%r>' % (self._Members(),)


def DetermineLocalPort(port_arg=0):
  if not port_arg:
    port_arg = portpicker.pick_unused_port()
  if not portpicker.is_port_free(port_arg):
    raise LocalPortUnavailableError('Local port [%d] is not available.' %
                                    port_arg)
  return port_arg


def _CloseLocalConnectionCallback(local_conn):
  """Callback function to close the local connection, if any."""
  # For test WebSocket connections, there is not a local socket connection.
  if local_conn:
    try:
      # Calling shutdown() first is needed to promptly notify the process on
      # the other side of the connection that it is closing. This allows that
      # other process, whether over TCP or stdin, to promptly terminate rather
      # that waiting for the next time that the process tries to send data.
      local_conn.shutdown(socket.SHUT_RDWR)
    except EnvironmentError:
      pass
    try:
      local_conn.close()
    except EnvironmentError:
      pass


def _GetAccessTokenCallback(credentials, lock):
  """Callback function to refresh credentials and return access token."""
  if not credentials:
    return None
  log.debug('credentials type for _GetAccessTokenCallback is [%s].',
            six.text_type(credentials))

  with lock:
    store.RefreshIfAlmostExpire(credentials)

    if creds.IsGoogleAuthCredentials(credentials):
      return credentials.token
    else:
      return credentials.access_token


def _SendLocalDataCallback(local_conn, data):
  # For test WebSocket connections, there is not a local socket connection.
  if local_conn:
    local_conn.send(data)


def _OpenLocalTcpSockets(local_host, local_port):
  """Attempt to open a local socket(s) listening on specified host and port."""
  open_sockets = []
  for res in socket.getaddrinfo(
      local_host, local_port, socket.AF_UNSPEC, socket.SOCK_STREAM, 0,
      socket.AI_PASSIVE):
    af, socktype, proto, unused_canonname, sock_addr = res
    try:
      s = socket.socket(af, socktype, proto)
    except socket.error:
      continue
    try:
      if not platforms.OperatingSystem.IsWindows():
        # This allows us to restart quickly on the same port. See b/213858080.
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
      s.bind(sock_addr)
      # Keep it large enough so it can handle burst of tunnel requests.
      s.listen(128)
      open_sockets.append(s)
    except EnvironmentError:
      try:
        s.close()
      except socket.error:
        pass

  if open_sockets:
    return open_sockets

  raise UnableToOpenPortError('Unable to open socket on port [%d].' %
                              local_port)


class _StdinSocket(object):
  """A wrapper around stdin/out that allows it to be treated like a socket.

  Does not implement all socket functions. And of the ones implemented, not all
  arguments/flags are supported. Once created, stdin should never be accessed by
  anything else.
  """

  class _StdinSocketMessage(object):
    """A class to wrap messages coming to the stdin socket for windows systems."""

    def __init__(self, message_type, data):
      self._type = message_type
      self._data = data

    def GetData(self):
      return self._data

    def GetType(self):
      return self._type

  class _EOFError(Exception):
    pass

  class _StdinClosedMessageType():
    pass

  class _ExceptionMessageType():
    pass

  class _DataMessageType():
    pass

  def __init__(self, ignore_windows_broken_pipe=False):

    self._stdin_closed = False
    self._ignore_windows_broken_pipe = ignore_windows_broken_pipe
    # Maximum number of bytes the thread should read.
    self._bufsize = utils.SUBPROTOCOL_MAX_DATA_FRAME_SIZE
    if platforms.OperatingSystem.IsWindows():
      # We will use this thread-safe queue to communicate with the input
      # reading thread.
      self._message_queue = queue.Queue()
      self._reading_thread = threading.Thread(
          target=self._ReadFromStdinAndEnqueueMessageWindows)
      self._reading_thread.daemon = True
      self._reading_thread.start()
    else:
      self._old_flags = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
      # Set up non-blocking mode to avoid getting stuck on read.
      fcntl.fcntl(sys.stdin, fcntl.F_SETFL, self._old_flags | os.O_NONBLOCK)

  def __del__(self):
    # We need to restore the flags, even when gcloud exits, nonblocking stdin
    # causes weird problems in that terminal, such as cat stops working.
    # This will happen if gcloud is running with stdin mode directly, then
    # killed between the set nonblocking and the restore. The user could fix
    # that by running bash and exiting it, or just closing the terminal.
    # If gcloud is running as an ssh ProxyCommand this problem doesn't happen.
    if not platforms.OperatingSystem.IsWindows():
      fcntl.fcntl(sys.stdin, fcntl.F_SETFL, self._old_flags)

  def send(self, data):  # pylint: disable=invalid-name
    files.WriteStreamBytes(sys.stdout, data)
    if not six.PY2:
      # WriteStreamBytes flushes python2 but not python3. Perhaps it should
      # be modified to also flush python3.
      sys.stdout.buffer.flush()
    return len(data)

  def recv(self, bufsize):  # pylint: disable=invalid-name
    """Receives data from stdin.

    Blocks until at least 1 byte is available.
    On Unix (but not Windows) this is unblocked by close() and shutdown(RD).
    On all platforms a signal handler triggering an exception will unblock this.
    This cannot be called by multiple threads at the same time.
    This function performs cleanups before returning, so killing gcloud while
    this is running should be avoided. Specifically RaisesKeyboardInterrupt
    should be in effect so that ctrl-c causes a clean exit with an exception
    instead of triggering gcloud's default os.kill().

    Args:
      bufsize: The maximum number of bytes to receive. Must be positive.
    Returns:
      The bytes received. EOF is indicated by b''.
    Raises:
      IOError: On low level errors.
    """

    if platforms.OperatingSystem.IsWindows():
      return self._RecvWindows(bufsize)
    else:
      return self._RecvUnix(bufsize)

  def close(self):  # pylint: disable=invalid-name
    # Closing stdin doesn't help, because it doesn't unblock read() calls.
    # Also it causes problems, such as segfaulting in python2 and blocking in
    # python3.
    self.shutdown(socket.SHUT_RD)

  def shutdown(self, how):  # pylint: disable=invalid-name
    # Shutting down read only (SHUT_RD)
    if how in (socket.SHUT_RDWR, socket.SHUT_RD):
      self._stdin_closed = True

      # For windows we will unblock the thread early
      if platforms.OperatingSystem.IsWindows():
        # We queue the message so that the recv loop can abort
        msg = self._StdinSocketMessage(self._StdinClosedMessageType, b'')
        self._message_queue.put(msg)

  def _ReadFromStdinAndEnqueueMessageWindows(self):
    """Reads data from stdin on Windows.

      This method will loop until stdin is closed. Should be executed in a
      separate thread to avoid blocking the main thread.
    """

    try:
      while not self._stdin_closed:
        # STD_INPUT_HANDLE is -10
        h = ctypes.windll.kernel32.GetStdHandle(-10)
        buf = ctypes.create_string_buffer(self._bufsize)
        number_of_bytes_read = wintypes.DWORD()
        ok = ctypes.windll.kernel32.ReadFile(
            h, buf, self._bufsize, ctypes.byref(number_of_bytes_read), None)
        if not ok:
          # There is a known issue with Windows IAP tunnel where stdin
          # will be closed with an ERROR_BROKEN_PIPE(109) on Windows. However,
          # the session was still closed properly. It is a short term fix to
          # ignore the broken pipe error on Windows.
          # See b/410043096 for more details and later proper fix.
          if (
              self._ignore_windows_broken_pipe
              and ctypes.windll.kernel32.GetLastError() == ERROR_BROKEN_PIPE
          ):
            self._stdin_closed = True
            msg = self._StdinSocketMessage(self._StdinClosedMessageType, b'')
            self._message_queue.put(msg)
            break
          raise socket.error(errno.EIO, 'stdin ReadFile failed')
        msg = buf.raw[:number_of_bytes_read.value]
        self._message_queue.put(self._StdinSocketMessage(self._DataMessageType,
                                                         msg))
    except Exception:  # pylint: disable=broad-except
      self._message_queue.put(
          self._StdinSocketMessage(self._ExceptionMessageType,
                                   sys.exc_info()))

  def _RecvWindows(self, bufsize):
    """Reads data from stdin on Windows.

    Args:
      bufsize: The maximum number of bytes to receive. Must be positive.
    Returns:
      The bytes received. EOF is indicated by b''.
    Raises:
      socket.error: On low level errors.
    """

    if bufsize != utils.SUBPROTOCOL_MAX_DATA_FRAME_SIZE:
      log.info('bufsize [%s] is not max_data_frame_size', bufsize)

    # We are using 1 second timeout here, which mean it can take up to 1
    # second for gcloud to realize it should exit. Lower timeout means more
    # cpu usage
    while not self._stdin_closed:
      try:
        msg = self._message_queue.get(timeout=1)
      except queue.Empty:
        # Timeout reached
        continue

      msg_type = msg.GetType()
      msg_data = msg.GetData()

      if msg_type is self._ExceptionMessageType:
        six.reraise(msg_data[0], msg_data[1], msg_data[2])

      if msg_type is self._StdinClosedMessageType:
        self._stdin_closed = True

      return msg_data
    # If stdin was closed we return an empty byte, so we can have a similar
    # behavior as the unix version
    return b''

  def _RecvUnix(self, bufsize):
    """Reads data from stdin on Unix.

    Args:
      bufsize: The maximum number of bytes to receive. Must be positive.
    Returns:
      The bytes received. EOF is indicated by b''. Once EOF has been indicated,
      will always indicate EOF.
    Raises:
      IOError: On low level errors.
    """

    # We don't except bufsize to be anything other than the max size of our
    # protocol. We will log it only for now to get telemetry if this
    # ever happens.
    if bufsize != utils.SUBPROTOCOL_MAX_DATA_FRAME_SIZE:
      log.info('bufsize [%s] is not max_data_frame_size', bufsize)

    if self._stdin_closed:
      return b''

    try:
      while not self._stdin_closed:
        # We have a timeout here because of b/197960494
        stdin_ready = select.select([sys.stdin], (), (),
                                    READ_FROM_STDIN_TIMEOUT_SECS)
        if not stdin_ready[0]:
          continue
        return self._ReadUnixNonBlocking(self._bufsize)
    except _StdinSocket._EOFError:
      self._stdin_closed = True
    return b''

  def _ReadUnixNonBlocking(self, bufsize):
    """Reads from stdin on Unix in a nonblocking manner.

    Args:
      bufsize: The maximum number of bytes to receive. Must be positive.
    Returns:
      The bytes read. b'' means no data is available.
    Raises:
      _StdinSocket._EOFError: to indicate EOF.
      IOError: On low level errors.
    """
    # In python 3, we need to read stdin in a binary way, not a text way to
    # read bytes instead of str. In python 2, binary mode vs text mode only
    # matters on Windows.
    try:
      if six.PY2:
        b = sys.stdin.read(bufsize)
      else:
        b = sys.stdin.buffer.raw.read(bufsize)
    except IOError as e:
      if e.errno == errno.EAGAIN or isinstance(e, io.BlockingIOError):
        # In python2, no nonblocking data available is indicated by raising
        # IOError with EAGAIN.
        # The online python3 documentation says BlockingIOError is raised when
        # no nonblocking data available. We handle that case in case it is ever
        # correct. BlockingIOError is a subclass of OSError which is identical
        # to IOError.
        return b''
      raise
    if b == b'':  # pylint: disable=g-explicit-bool-comparison
      # In python 2 and 3, EOF is indicated by returning b''.
      raise _StdinSocket._EOFError

    if b is None:
      # Regardless of what the online python3 documentation says, it actually
      # returns None to indicate no nonblocking data available.
      b = b''
    return b


class SecurityGatewayTunnelHelper(object):
  """Helper class for starting a Security Gateaway tunnel."""

  def __init__(self, args, project, region, security_gateway, host, port,
               use_dest_group=False):
    # Re-use the same args as IAP to prevent adding more flags than necessary.
    self._tunnel_url_override = args.iap_tunnel_url_override
    self._ignore_certs = args.iap_tunnel_insecure_disable_websocket_cert_check

    self._project = project
    self._region = region
    self._security_gateway = security_gateway
    self._host = host
    self._port = port

    self._use_dest_group = use_dest_group

    self._shutdown = False

    self._credential = store.LoadIfEnabled()
    self._credential_lock = threading.Lock()

  def _InitiateConnection(self, local_conn,
                          get_access_token_callback, user_agent):
    del user_agent  # Unused.
    sg_tunnel_target = self._GetTargetInfo()
    new_sg_tunnel = sg_tunnel.SecurityGatewayTunnel(
        sg_tunnel_target,
        get_access_token_callback,
        functools.partial(_SendLocalDataCallback, local_conn),
        functools.partial(_CloseLocalConnectionCallback, local_conn),
        self._ignore_certs)
    new_sg_tunnel.InitiateConnection()
    return new_sg_tunnel

  def _GetTargetInfo(self):
    proxy_info = http_proxy.GetHttpProxyInfo()
    if callable(proxy_info):
      proxy_info = proxy_info(method='https')
    return sg_utils.SecurityGatewayTargetInfo(
        project=self._project,
        region=self._region,
        security_gateway=self._security_gateway,
        host=self._host,
        port=self._port,
        url_override=self._tunnel_url_override,
        proxy_info=proxy_info,
        use_dest_group=self._use_dest_group,
    )

  def RunReceiveLocalData(self, local_conn, socket_address, user_agent,
                          conn_id=-1):
    """Receive data from provided local connection and send over HTTP CONNECT.

    Args:
      local_conn: A socket or _StdinSocket representing the local connection.
      socket_address: A verbose loggable string describing where conn is
        connected to.
      user_agent: The user_agent of this connection
      conn_id: The id of the connection.
    """
    del conn_id  # Unused.
    sg_conn = None
    try:
      sg_conn = self._InitiateConnection(
          local_conn,
          functools.partial(
              _GetAccessTokenCallback, self._credential, self._credential_lock
          ),
          user_agent,
      )
      while not (self._shutdown or sg_conn.ShouldStop()):
        data = local_conn.recv(utils.SUBPROTOCOL_MAX_DATA_FRAME_SIZE)
        if not data:
          log.warning('Local connection [%s] has closed.', socket_address)
          break
        sg_conn.Send(data)
    except socket.error as e:
      log.error('Error while transmitting local connection [%s]: %s ',
                socket_address, e)
    finally:
      log.info('Terminating connection from local connection: [%s]',
               socket_address)
      if local_conn:
        local_conn.shutdown(socket.SHUT_RD)
        local_conn.close()
      if sg_conn:
        sg_conn.Close()
        log.debug('Connection [%s] closed.', socket_address)

  def Close(self):
    # This is expected to be called from a separate thread than the one running
    # RunReceiveLocalData.
    self._shutdown = True


class IAPWebsocketTunnelHelper(object):
  """Helper class for starting an IAP WebSocket tunnel."""

  def __init__(self, args, project,
               zone=None, instance=None, interface=None, port=None,
               region=None, network=None, host=None, dest_group=None):
    self._project = project
    self._iap_tunnel_url_override = args.iap_tunnel_url_override
    self._ignore_certs = args.iap_tunnel_insecure_disable_websocket_cert_check

    self._zone = zone
    self._instance = instance
    self._interface = interface
    self._port = port
    self._region = region
    self._network = network
    self._host = host
    self._dest_group = dest_group

    self._shutdown = False

    self._credential = store.LoadIfEnabled()
    self._credential_lock = threading.Lock()

  def Close(self):
    self._shutdown = True

  def _InitiateConnection(self, local_conn, get_access_token_callback,
                          user_agent, conn_id=-1):
    tunnel_target = self._GetTunnelTargetInfo()
    new_websocket = iap_tunnel_websocket.IapTunnelWebSocket(
        tunnel_target, get_access_token_callback,
        functools.partial(_SendLocalDataCallback, local_conn),
        functools.partial(_CloseLocalConnectionCallback, local_conn),
        user_agent, ignore_certs=self._ignore_certs,
        conn_id=conn_id)
    new_websocket.InitiateConnection()
    return new_websocket

  def _GetTunnelTargetInfo(self):
    proxy_info = http_proxy.GetHttpProxyInfo()
    if callable(proxy_info):
      proxy_info = proxy_info(method='https')
    return utils.IapTunnelTargetInfo(
        project=self._project,
        zone=self._zone,
        instance=self._instance,
        interface=self._interface,
        port=self._port,
        url_override=self._iap_tunnel_url_override,
        proxy_info=proxy_info,
        region=self._region,
        network=self._network,
        host=self._host,
        dest_group=self._dest_group,
        cloud_run_args=None,
    )

  def RunReceiveLocalData(self, conn, socket_address, user_agent, conn_id=0):
    """Receive data from provided local connection and send over WebSocket.

    Args:
      conn: A socket or _StdinSocket representing the local connection.
      socket_address: A verbose loggable string describing where conn is
        connected to.
      user_agent: The user_agent of this connection
      conn_id: Id of the connection.
    """
    websocket_conn = None
    try:
      websocket_conn = self._InitiateConnection(
          conn,
          functools.partial(
              _GetAccessTokenCallback, self._credential, self._credential_lock
          ),
          user_agent,
          conn_id=conn_id,
      )
      while not self._shutdown:
        data = conn.recv(utils.SUBPROTOCOL_MAX_DATA_FRAME_SIZE)
        if not data:
          # When we recv an EOF, we notify the websocket_conn of it, then we
          # wait for all data to send before returning.
          websocket_conn.LocalEOF()
          log.debug('[%d] Received local EOF, closing connection', conn_id)
          if not websocket_conn.WaitForAllSent():
            log.warning('[%d] Failed to send all data from [%s].',
                        conn_id, socket_address)
          break
        websocket_conn.Send(data)
    except (Exception, exceptions.Error) as e:  # pylint: disable=broad-exception-caught
      log.exception('[%d] Error during local connection to [%s]: %s', conn_id,
                    socket_address, e)
    finally:
      if self._shutdown:
        log.info('[%d] Terminating connection to [%s].',
                 conn_id, socket_address)
      else:
        log.info('[%d] Client closed connection from [%s].',
                 conn_id, socket_address)
      try:
        conn.close()
      except EnvironmentError:
        pass
      try:
        if websocket_conn:
          websocket_conn.Close()
      except (EnvironmentError, exceptions.Error):
        pass


class IapTunnelProxyServerHelper():
  """Proxy server helper listens on a port for new local connections."""

  def __init__(self, local_host, local_port,
               should_test_connection, tunneler):
    self._tunneler = tunneler
    self._local_host = local_host
    self._local_port = local_port
    self._should_test_connection = should_test_connection
    self._server_sockets = []
    self._connections = []
    self._total_connections = 0

  def __del__(self):
    self._CloseServerSockets()

  def Run(self):
    """Start accepting connections."""
    if self._should_test_connection:
      try:
        self._TestConnection()
      except iap_tunnel_websocket.ConnectionCreationError as e:
        raise iap_tunnel_websocket.ConnectionCreationError(
            'While checking if a connection can be made: %s' % six.text_type(e))
    self._server_sockets = _OpenLocalTcpSockets(self._local_host,
                                                self._local_port)
    log.out.Print('Listening on port [%d].' % self._local_port)

    try:
      with execution_utils.RaisesKeyboardInterrupt():
        while True:
          self._connections.append(self._AcceptNewConnection())
          # To fix b/189195317, we will need to erase the reference of dead
          # tasks.
          self._CleanDeadClientConnections()
    except KeyboardInterrupt:
      log.info('Keyboard interrupt received.')
    finally:
      self._CloseServerSockets()

    self._tunneler.Close()
    self._CloseClientConnections()
    log.status.Print('Server shutdown complete.')

  def _TestConnection(self):
    """Test if a connection can be made to the requested endpoint."""
    log.status.Print('Testing if tunnel connection works.')
    user_agent = transport.MakeUserAgentString()
    # pylint: disable=protected-access
    conn = self._tunneler._InitiateConnection(
        None,
        functools.partial(
            _GetAccessTokenCallback,
            self._tunneler._credential,
            threading.Lock(),
        ),
        user_agent,
    )
    # pylint: enable=protected-access
    conn.Close()

  def _AcceptNewConnection(self):
    """Accept a new socket connection and start a new WebSocket tunnel."""
    # Python socket accept() on Windows does not get interrupted by ctrl-c
    # To work around that, use select() with a timeout before the accept()
    # which allows for the ctrl-c to be noticed and abort the process as
    # expected.
    ready_sockets = [()]
    while not ready_sockets[0]:
      # 0.2 second timeout
      ready_sockets = select.select(self._server_sockets, (), (), 0.2)

    ready_read_sockets = ready_sockets[0]
    conn, socket_address = ready_read_sockets[0].accept()
    new_thread = threading.Thread(target=self._HandleNewConnection,
                                  args=(conn, socket_address,
                                        self._total_connections))
    new_thread.daemon = True
    new_thread.start()
    self._total_connections += 1
    return new_thread, conn

  def _CloseServerSockets(self):
    log.debug('Stopping server.')
    try:
      for server_socket in self._server_sockets:
        server_socket.close()
    except EnvironmentError:
      pass

  def _CloseClientConnections(self):
    """Close client connections that seem to still be open."""
    if self._connections:
      close_count = 0
      for client_thread, conn in self._connections:
        if client_thread.is_alive():
          close_count += 1
          try:
            conn.close()
          except EnvironmentError:
            pass
      if close_count:
        log.status.Print('Closed [%d] local connection(s).' % close_count)

  def _CleanDeadClientConnections(self):
    """Erase reference to dead connections so they can be garbage collected."""
    conn_still_alive = []
    if self._connections:
      dead_connections = 0
      for client_thread, conn in self._connections:
        if not client_thread.is_alive():
          dead_connections += 1
          try:
            conn.close()
          except EnvironmentError:
            pass
          del conn
          del client_thread
        else:
          conn_still_alive.append([client_thread, conn])
      if dead_connections:
        log.debug('Cleaned [%d] dead connection(s).' % dead_connections)
        self._connections = conn_still_alive
      # We run GC mostly for windows platforms, where it seems GC is not
      # collecting memory quick enough. For linux platforms, this is needed only
      # to immediately clean the memory we freed above.
      gc.collect(2)
      log.debug('connections alive: [%d]' % len(self._connections))

  def _HandleNewConnection(self, conn, socket_address, conn_id):
    try:
      user_agent = transport.MakeUserAgentString()
      self._tunneler.RunReceiveLocalData(conn, repr(socket_address),
                                         user_agent, conn_id=conn_id)
    except EnvironmentError as e:
      log.info('Socket error [%s] while receiving from client.',
               six.text_type(e))
    except:  # pylint: disable=bare-except
      log.exception('Error while receiving from client.')


class IapTunnelStdinHelper():
  """Facilitates a connection that gets local data from stdin."""

  def __init__(
      self,
      tunneler,
      with_graceful_shutdown=False,
      ignore_windows_broken_pipe=False,
  ):
    self._tunneler = tunneler
    self._with_graceful_shutdown = with_graceful_shutdown
    self._ignore_windows_broken_pipe = ignore_windows_broken_pipe

  def Run(self):
    """Executes the tunneling of data."""
    local_conn = _StdinSocket(
        ignore_windows_broken_pipe=self._ignore_windows_broken_pipe
    )

    def _HandleTerminationSignal(signum, frame):
      del signum, frame
      # Closing the stdin wrapper causes the read loop to hit EOF,
      # which triggers the finally block to flush the WebSocket CLOSE frame.
      local_conn.close()

    original_sig_handlers = {}
    if self._with_graceful_shutdown:
      for sig_name in _TERMINATION_SIGNALS:
        if hasattr(signal, sig_name):
          sig = getattr(signal, sig_name)
          original_sig_handlers[sig] = signal.signal(
              sig, _HandleTerminationSignal
          )

    try:
      with execution_utils.RaisesKeyboardInterrupt():

        # Fetching user agent before we start the read loop, because the agent
        # fetch will call sys.stdin.isatty, which is blocking if there is a read
        # waiting for data in the stdin. This only affects MacOs + python 2.7.
        user_agent = transport.MakeUserAgentString()

        self._tunneler.RunReceiveLocalData(local_conn, 'stdin', user_agent)
    except KeyboardInterrupt:
      log.info('Keyboard interrupt received.')
    finally:
      if self._with_graceful_shutdown:
        for sig, handler in original_sig_handlers.items():
          signal.signal(sig, handler)



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Restores a Compute Disk Backup."""



from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.backupdr import util
from googlecloudsdk.api_lib.backupdr.backups import BackupsClient
from googlecloudsdk.api_lib.backupdr.backups import DiskRestoreConfig
from googlecloudsdk.api_lib.util import exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as calliope_exceptions
from googlecloudsdk.command_lib.backupdr import flags
from googlecloudsdk.command_lib.backupdr.restore import disk_flags
from googlecloudsdk.core import log


@base.DefaultUniverseOnly
@base.ReleaseTracks(
    base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA, base.ReleaseTrack.GA
)
class Disk(base.Command):
  """Restores a Compute Disk Backup."""

  detailed_help = {
      'BRIEF': 'Restores the specified backup',
      'DESCRIPTION': '{description}',
      'EXAMPLES': """\
        To restore a backup `sample-backup` in project `sample-project` and location `us-central1`,
        with `sample-data-store` and `sample-backup-vault`, and additional target properties, run:

          $ {command} sample-backup --project=sample-project --location=us-central1 --backup-vault=sample-backup-vault --data-source=sample-data-source --<target-properties>
        """,
  }

  @staticmethod
  def Args(parser):
    """Specifies additional command flags.

    Args:
      parser: argparse.Parser: Parser object for command line inputs.
    """
    base.ASYNC_FLAG.AddToParser(parser)
    base.ASYNC_FLAG.SetDefault(parser, True)
    flags.AddBackupResourceArg(
        parser, 'The backup of a resource to be restored.'
    )

    disk_flags.AddNameArg(parser)
    disk_flags.AddSourceInstanceDiskArgs(parser)
    disk_flags.AddTargetZoneArg(parser, False)
    disk_flags.AddTargetRegionArg(parser, False)
    disk_flags.AddTargetProjectArg(parser)
    disk_flags.AddReplicaZonesArg(parser, False)
    disk_flags.AddGuestOsFeaturesArgs(parser, False)
    disk_flags.AddDescriptionArg(parser, False)
    disk_flags.AddLicensesArg(parser, False)
    disk_flags.AddLabelsArg(parser, False)
    disk_flags.AddTypeArg(parser, False)
    disk_flags.AddAccessModeArg(parser, False)
    disk_flags.AddProvisionedIopsArg(parser, False)
    disk_flags.AddProvisionedThroughputArg(parser, False)
    disk_flags.AddArchitectureArg(parser, False)
    disk_flags.AddStoragePoolArg(parser, False)
    disk_flags.AddSizeArg(parser, False)
    disk_flags.AddConfidentialComputeArg(parser, False)
    disk_flags.AddResourcePoliciesArg(parser, False)
    disk_flags.AddKmsKeyArg(parser, False)
    disk_flags.AddClearEncryptionKeyArg(parser)
    flags.AddUseProjectServiceAccountFlag(parser)

  def _ParseResourcePolicies(self, resource_policies, project, zone):
    """Parses the resource policies flag."""
    resource_policy_uris = []
    for policy in resource_policies:
      if not policy.startswith('projects/'):
        region = zone.rsplit('-', 1)[0]
        policy = 'projects/{}/regions/{}/resourcePolicies/{}'.format(
            project, region, policy
        )
      resource_policy_uris.append(policy)
    return resource_policy_uris

  def Run(self, args):
    """Constructs and sends request.

    Args:
      args: argparse.Namespace, An object that contains the values for the
        arguments specified in the .Args() method.

    Returns:
      ProcessHttpResponse of the request made.
    """
    api_version = util.GetApiVersion(self.ReleaseTrack())
    client = BackupsClient(api_version=api_version)
    is_async = args.async_

    backup = args.CONCEPTS.backup.Parse()
    restore_config = DiskRestoreConfig()
    restore_config['Name'] = args.name
    restore_config['TargetProject'] = args.target_project
    if args.target_zone:
      restore_config['TargetZone'] = args.target_zone
    if args.target_region:
      restore_config['TargetRegion'] = args.target_region
    if args.target_region and not args.replica_zones:
      raise calliope_exceptions.RequiredArgumentException(
          'replica_zones',
          'Replica zones are required for regional disk restore.',
      )
    if args.replica_zones:
      restore_config['ReplicaZones'] = args.replica_zones
    if args.guest_os_features:
      restore_config['GuestOsFeatures'] = args.guest_os_features
    if args.licenses:
      restore_config['Licenses'] = args.licenses
    if args.description:
      restore_config['Description'] = args.description
    if args.type:
      restore_config['Type'] = args.type
    if args.access_mode:
      restore_config['AccessMode'] = args.access_mode
    if args.provisioned_iops:
      restore_config['ProvisionedIops'] = args.provisioned_iops
    if args.provisioned_throughput:
      restore_config['ProvisionedThroughput'] = args.provisioned_throughput
    if args.architecture:
      restore_config['Architecture'] = args.architecture
    if args.storage_pool:
      restore_config['StoragePool'] = args.storage_pool
    if args.size:
      restore_config['Size'] = args.size
    if args.kms_key:
      restore_config['KmsKey'] = args.kms_key
    if args.labels:
      restore_config['Labels'] = args.labels
    if args.confidential_compute:
      restore_config['ConfidentialCompute'] = args.confidential_compute
    if args.resource_policies:
      restore_config['ResourcePolicies'] = self._ParseResourcePolicies(
          args.resource_policies, args.target_project, args.target_zone
      )
    if args.source_instance_boot_disk:
      restore_config['SourceInstanceBootDisk'] = args.source_instance_boot_disk
    if args.source_instance_disk_device_name:
      restore_config['SourceInstanceDiskDeviceName'] = (
          args.source_instance_disk_device_name
      )
    if args.use_project_service_account:
      restore_config['UseProjectServiceAccount'] = (
          args.use_project_service_account
      )
    if (
        hasattr(args, 'clear_encryption_key')
        and args.clear_encryption_key
    ):
      restore_config['ClearOverridesFieldMask'] = (
          'disk_restore_properties.disk_encryption_key'
      )
    try:
      operation = client.RestoreDisk(backup, restore_config)
    except apitools_exceptions.HttpError as e:
      raise exceptions.HttpException(e, util.HTTP_ERROR_FORMAT) from e

    if is_async:
      log.RestoredResource(
          backup.Name(),
          kind='backup',
          is_async=True,
          details=util.ASYNC_OPERATION_MESSAGE.format(operation.name),
      )
      return operation

    # For sync operation
    return client.WaitForOperation(
        operation_ref=client.GetOperationRef(operation),
        message=(
            'Restoring backup'
            ' [{}].'
            ' (This operation could take upto 15 minutes.)'.format(
                backup.Name()
            )
        ),
        has_result=False,
    )
